import ast
import logging
import os
from collections.abc import Iterable
from typing import Any

from boa3.internal import constants
from boa3.internal.analyser import asthelper
from boa3.internal.analyser.astanalyser import IAstAnalyser
from boa3.internal.analyser.importanalyser import ImportAnalyser
from boa3.internal.analyser.model.ManifestSymbol import ManifestSymbol
from boa3.internal.analyser.model.functionarguments import FunctionArguments
from boa3.internal.analyser.model.optimizer import UndefinedType
from boa3.internal.analyser.model.symbolscope import SymbolScope
from boa3.internal.exception import CompilerError, CompilerWarning
from boa3.internal.model.builtin.builtin import Builtin
from boa3.internal.model.builtin.compile_time.neometadatatype import MetadataTypeSingleton
from boa3.internal.model.builtin.decorator import ContractDecorator
from boa3.internal.model.builtin.decorator.builtindecorator import IBuiltinDecorator
from boa3.internal.model.builtin.interop.runtime import ScriptContainerProperty
from boa3.internal.model.builtin.interop.runtime import NotifyMethod
from boa3.internal.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.internal.model.callable import Callable
from boa3.internal.model.decorator import IDecorator
from boa3.internal.model.event import Event
from boa3.internal.model.expression import IExpression
from boa3.internal.model.imports.importsymbol import BuiltinImport, Import
from boa3.internal.model.imports.package import Package
from boa3.internal.model.method import Method
from boa3.internal.model.module import Module
from boa3.internal.model.property import Property
from boa3.internal.model.symbol import ISymbol
from boa3.internal.model.type.annotation.metatype import MetaType
from boa3.internal.model.type.annotation.uniontype import UnionType
from boa3.internal.model.type.classes.classscope import ClassScope
from boa3.internal.model.type.classes.classtype import ClassType
from boa3.internal.model.type.classes.contractinterfaceclass import ContractInterfaceClass
from boa3.internal.model.type.classes.pythonclass import PythonClass
from boa3.internal.model.type.classes.userclass import UserClass
from boa3.internal.model.type.collection.icollection import ICollectionType as Collection
from boa3.internal.model.type.collection.sequence.sequencetype import SequenceType
from boa3.internal.model.type.primitive.primitivetype import PrimitiveType
from boa3.internal.model.type.type import IType, Type
from boa3.internal.model.variable import Variable
from boa3.sc.compiletime import NeoMetadata


class ModuleAnalyser(IAstAnalyser, ast.NodeVisitor):
    """
    This class is responsible for mapping the locals of the functions and modules

    The methods with the name starting with 'visit_' are implementations of methods from the :class:`NodeVisitor` class.
    These methods are used to walk through the Python abstract syntax tree.

    :ivar modules: a dictionary that maps each module with its name. Empty by default.
    :ivar symbols: a dictionary that maps the global symbols.
    """

    def __init__(self, analyser, symbol_table: dict[str, ISymbol],
                 filename: str = None, root_folder: str = None,
                 analysed_files: dict | None = None,
                 import_stack: list[str] | None = None,
                 log: bool = False,
                 fail_fast: bool = True):
        super().__init__(analyser.ast_tree, filename, root_folder, log, fail_fast)
        self.modules: dict[str, Module] = {}
        self.symbols: dict[str, ISymbol] = symbol_table

        from boa3.internal.analyser.analyser import Analyser
        if isinstance(analysed_files, dict):
            for file_path, file_analyser in analysed_files.copy().items():
                fixed_path = file_path.replace(os.sep, constants.PATH_SEPARATOR)
                if file_path != fixed_path:
                    analysed_files.pop(file_path)
                    analysed_files[fixed_path] = file_analyser
        else:
            analysed_files = {}

        analysed_files[filename.replace(os.sep, constants.PATH_SEPARATOR)] = analyser
        self._analysed_files: dict[str, Analyser] = analysed_files

        if isinstance(import_stack, list):
            import_stack = [file_path.replace(os.sep, constants.PATH_SEPARATOR)
                            if isinstance(file_path, str) else file_path
                            for file_path in import_stack]
        else:
            import_stack = []
        self._import_stack: list[str] = import_stack

        self._builtin_functions_to_visit: dict[str, IBuiltinMethod] = {}
        self._current_module: Module = None
        self._current_class: UserClass = None
        self._current_method: Method = None
        self._current_event: Event = None

        self._deploy_method: Method | None = None

        self._annotated_variables: list[str] = []
        self._global_assigned_variables: list[str] = []
        self._scope_stack: list[SymbolScope] = []

        self._metadata: NeoMetadata = None
        self._metadata_node: ast.AST = ast.parse('')
        self._manifest_symbols: dict[tuple[ManifestSymbol, str, int], Callable] = {}
        self.imported_nodes: list[ast.AST] = []

        if self.filename:
            self._tree.filename = self.filename
        self.analyse_visit(self._tree)

        analyser.metadata = self._metadata if self._metadata is not None else NeoMetadata()

    @property
    def _current_scope(self) -> Method | Module | UserClass | None:
        """
        Returns the scope that is currently being analysed

        :return: the current scope. Return None if it is the global scope
        :rtype: Method or Module or None
        """
        if self._current_method is not None:
            return self._current_method
        if self._current_class is not None:
            return self._current_class
        return self._current_module

    @property
    def _current_symbol_scope(self) -> SymbolScope | None:
        if len(self._scope_stack) > 0:
            return self._scope_stack[-1]
        else:
            return None

    @property
    def global_symbols(self) -> dict[str, ISymbol]:
        """
        Returns the symbols in global scope

        :return: the symbol table of the global scope
        """
        global_symbols: dict[str, ISymbol] = {}

        global_symbols.update(self.symbols)
        for mod in self.modules.values():
            global_symbols.update(mod.symbols)

        return global_symbols

    @property
    def analysed_files(self) -> dict[str, Any]:
        return self._analysed_files.copy()

    def __include_variable(self,
                           var_id: str, var_type_id: str | IType,
                           source_node: ast.AST,
                           var_enumerate_type: IType = Type.none,
                           assignment: bool = True,
                           literal_value: Any = asthelper.INVALID_NODE_RESULT):
        """
        Includes the variable in the symbol table if the id was not used

        :param var_id: variable id
        :param var_type_id: variable type id
        :type var_type_id: str or IType
        :param var_enumerate_type: variable value type id if var_type_id is a SequenceType
        """
        if not isinstance(var_id, str) and isinstance(var_id, Iterable):
            variables = [var_name.id if isinstance(var_name, ast.Name) else self.visit(var_name).id
                         for var_name in source_node.targets[0].elts]
            var_types = self.visit(source_node.value)
        else:
            variables = [var_id]
            var_types = [var_type_id]

        for x in range(min(len(variables), len(var_types))):
            var_id, var_type_id = variables[x], var_types[x]

            outer_symbol = self.get_symbol(var_id)
            first_assign = outer_symbol is None
            if var_id in self._current_symbol_scope.symbols:
                if hasattr(outer_symbol, 'set_is_reassigned'):
                    is_module_scope = isinstance(self._current_scope, Module)
                    if not is_module_scope:
                        outer_symbol.set_is_reassigned()
                    elif isinstance(source_node, ast.AugAssign):
                        # augmented assignments of global variables shouldn't be evaluated in the module analyser
                        outer_symbol.set_is_reassigned()
                    self.__set_source_origin(source_node, is_module_scope)
            else:
                if (not isinstance(source_node, ast.Global) and
                        (not hasattr(source_node, 'targets') or not isinstance(source_node.targets[x], ast.Subscript))):
                    if outer_symbol is not None:
                        self._log_warning(
                            CompilerWarning.NameShadowing(source_node.lineno, source_node.col_offset, outer_symbol, var_id)
                        )
                        first_assign = True

                var_type = None
                if isinstance(var_type_id, SequenceType):
                    var_type = var_type_id
                    var_enumerate_type = self.get_enumerate_type(var_type)
                elif isinstance(var_type_id, IType):
                    var_type = var_type_id

                # when setting a None value to a variable, set the variable as any type
                if var_type is Type.none:
                    var_type = Type.any

                if isinstance(var_type, IType) or var_type is None:
                    # if type is None, the variable type depends on the type of a expression
                    if isinstance(source_node, ast.Global):
                        var = outer_symbol
                    else:
                        if isinstance(var_type, SequenceType) and not Type.tuple.is_type_of(var_type):
                            var_type = var_type.build_collection(var_enumerate_type)
                        var = Variable(var_type, origin_node=source_node)

                    if first_assign and isinstance(Type.get_type(literal_value), PrimitiveType):
                        var.set_initial_assign(literal_value)

                    self._current_symbol_scope.include_symbol(var_id, var)
                    if isinstance(source_node, ast.AnnAssign):
                        self._annotated_variables.append(var_id)
            if hasattr(self._current_scope, 'assign_variable') and assignment:
                self._global_assigned_variables.append(var_id)

    def __include_callable(self, callable_id: str, callable: Callable):
        """
        Includes the method in the symbol table if the id was not used

        :param callable_id: method id
        :param callable: method to be included
        """
        if ((self._current_scope is self._current_class or callable_id not in self._current_scope.symbols)
                and hasattr(self._current_scope, 'include_callable')):
            already_exists = not self._current_scope.include_callable(callable_id, callable)
        else:
            symbol = self.get_symbol(callable_id)
            already_exists = symbol is not None and not isinstance(symbol, IBuiltinMethod)

        if already_exists:
            self._log_error(CompilerError.DuplicatedIdentifier(callable.origin.lineno,
                                                               callable.origin.col_offset,
                                                               callable_id))

        if callable.is_public:
            # check if the external name + argument number is unique
            manifest_name = callable.external_name if callable.external_name is not None else callable_id
            manifest_id = (ManifestSymbol.get_manifest_symbol(callable), manifest_name, len(callable.args))

            if manifest_id in self._manifest_symbols:
                self._log_error(CompilerError.DuplicatedManifestIdentifier(callable.origin.lineno,
                                                                           callable.origin.col_offset,
                                                                           manifest_name, len(callable.args)
                                                                           ))
            else:
                self._manifest_symbols[manifest_id] = callable

    def __include_class_variable(self, cl_var_id: str, cl_var: Variable):
        """
        Includes the class variable in the current class

        :param cl_var_id: variable name
        :param cl_var: variable to be included
        """
        if cl_var_id not in self._current_scope.class_variables:
            self._current_class.include_symbol(cl_var_id, cl_var, ClassScope.CLASS)

    def __set_source_origin(self, source_node: ast.AST, scope_is_correct: bool = True):
        if scope_is_correct or self._current_scope == self._deploy_method:
            source_node.origin = self._tree

    def get_symbol(self, symbol_id: str,
                   is_internal: bool = False,
                   check_raw_id: bool = False,
                   origin_node: ast.AST = None) -> ISymbol | None:
        for scope in reversed(self._scope_stack):
            if symbol_id in scope.symbols:
                return scope.symbols[symbol_id]

            if check_raw_id:
                found_symbol = self._search_by_raw_id(symbol_id, list(scope.symbols.values()))
                if found_symbol is not None:
                    return found_symbol

        if symbol_id in self._current_scope.symbols:
            # the symbol exists in the local scope
            return self._current_scope.symbols[symbol_id]
        elif symbol_id in self._current_module.symbols:
            # the symbol exists in the module scope
            return self._current_module.symbols[symbol_id]

        if check_raw_id:
            found_symbol = self._search_by_raw_id(symbol_id, list(self._current_scope.symbols.values()))
            if found_symbol is not None:
                return found_symbol

            found_symbol = self._search_by_raw_id(symbol_id, list(self._current_module.symbols.values()))
            if found_symbol is not None:
                return found_symbol

        return super().get_symbol(symbol_id, is_internal, check_raw_id, origin_node)

    def get_annotation(self, value: Any, use_metatype: bool = False, accept_none: bool = False) -> IType | None:
        if not isinstance(value, ast.AST):
            return None

        annotation_type = self.get_type(value, use_metatype)
        if not isinstance(annotation_type, PythonClass):
            return annotation_type
        if hasattr(value, 'value') and value.value is None and annotation_type is Type.none:
            return annotation_type

        if isinstance(value, (ast.Constant, ast.NameConstant, ast.List, ast.Tuple, ast.Dict, ast.Set)):
            # annotated types should only accept types
            return None
        return annotation_type

    def _check_annotation_type(self, node: ast.AST, origin_node: ast.AST | None = None):
        if node is None:
            return

        if origin_node is None:
            origin_node = node

        if self.get_annotation(node) is None:
            actual_type = self.get_type(node)
            self._log_error(
                CompilerError.MismatchedTypes(
                    origin_node.lineno, origin_node.col_offset,
                    expected_type_id=type.__name__,
                    actual_type_id=actual_type.identifier
                ))

    # region Log

    def _log_import(self, import_from: str):
        if self._log:
            logging.getLogger(constants.BOA_LOGGING_NAME).info("Importing '{0}'\t <{1}>".format(import_from, self.filename))

    def _log_unresolved_import(self, origin_node: ast.AST, import_id: str):
        if self._log:
            self._log_error(
                CompilerError.UnresolvedReference(
                    line=origin_node.lineno,
                    col=origin_node.col_offset,
                    symbol_id=import_id
                )
            )

    # endregion

    # region Metadata

    def _read_metadata_object(self, function: ast.FunctionDef):
        """
        Gets the metadata object defined in this function

        :param function:
        """
        if self._metadata is not None:
            # metadata function has been defined already
            self._log_warning(
                CompilerWarning.RedeclaredSymbol(
                    line=function.lineno, col=function.col_offset,
                    symbol_id="metadata"
                )
            )
        # this function must have a return and no arguments
        elif len(function.args.args) != 0:
            self._log_error(
                CompilerError.UnexpectedArgument(
                    line=function.lineno, col=function.col_offset
                )
            )
        elif not any(isinstance(stmt, ast.Return) for stmt in function.body):
            self._log_error(
                CompilerError.MissingReturnStatement(
                    line=function.lineno, col=function.col_offset,
                    symbol_id=function.name
                )
            )
        else:
            function.returns = None
            function.decorator_list = []

            imports: list[ast.AST] = []
            other_instructions: list[ast.AST] = []
            for node in self._tree.body:
                if node == function:
                    # metadata function must be right after all the imports, so it executes correctly
                    continue

                if isinstance(node, (ast.ImportFrom, ast.Import)):
                    imports.append(node)
                else:
                    other_instructions.append(node)

            module: ast.Module = ast.parse('')
            module.body = imports + [function] + other_instructions
            ast.copy_location(module, function)
            namespace = {}

            import sys
            file_dir = os.path.abspath(os.path.dirname(self.filename))
            sys_path = sys.path.copy()

            try:
                from boa3.internal import utils
                if os.path.abspath(self.root_folder) != file_dir:
                    sc_paths = [os.path.abspath(self.root_folder), file_dir]
                else:
                    sc_paths = [file_dir]

                sys.path = sc_paths + utils.list_inner_packages(file_dir) + sys_path

                # executes the function
                code = compile(module, filename='<boa3>', mode='exec')
                exec(code, namespace)
            except ModuleNotFoundError:
                # will fail if any imports can't be executed
                # in this case, the error is already logged
                return
            except BaseException as inner_exception:
                # reordering the module tree may raise unexpected exceptions
                # ignore if it has generated the metadata function
                if function.name not in namespace:
                    raise inner_exception
            finally:
                sys.path = sys_path

            obj: Any = namespace[function.name]()
            node: ast.AST = function.body[-1] if len(function.body) > 0 else function
            # return must be a NeoMetadata object
            if not isinstance(obj, NeoMetadata):
                obj_type = self.get_type(obj).identifier if self.get_type(obj) is not Type.any else type(obj).__name__
                self._log_error(
                    CompilerError.MismatchedTypes(
                        line=node.lineno, col=node.col_offset,
                        expected_type_id=NeoMetadata.__name__,
                        actual_type_id=obj_type
                    )
                )
                return

            # validate if the extras field can be converted to json
            try:
                import json
                json.dumps(obj.extras)
            except BaseException as e:
                self._log_error(
                    CompilerError.InvalidType(
                        line=node.lineno, col=node.col_offset,
                        symbol_id=str(e)
                    )
                )
                return

            # validates the metadata attributes types
            attributes: dict[str, Any] = {attr: value
                                          for attr, value in dict(obj.__dict__).items()
                                          if attr in Builtin.metadata_fields}
            if any(not isinstance(value, Builtin.metadata_fields[attr]) for attr, value in attributes.items()):
                for expected, actual in [(Builtin.metadata_fields[attr], type(v_type))
                                         for attr, v_type in attributes.items()
                                         if not isinstance(v_type, Builtin.metadata_fields[attr])]:
                    if isinstance(expected, Iterable):
                        expected_id = 'Union[{0}]'.format(', '.join([tpe.__name__ for tpe in expected]))
                    elif hasattr(expected, '__name__'):
                        expected_id = expected.__name__
                    else:
                        expected_id = str(expected)

                    self._log_error(
                        CompilerError.MismatchedTypes(
                            line=node.lineno, col=node.col_offset,
                            expected_type_id=expected_id,
                            actual_type_id=actual.__name__
                        )
                    )
            else:
                # if the function was defined correctly, sets the metadata object of the smart contract
                self._metadata = obj
                self._metadata_node = function  # for error messages only

    # endregion

    # region AST

    def visit_ImportFrom(self, import_from: ast.ImportFrom):
        """
        Includes methods and variables from other modules into the current scope

        :param import_from:
        """
        self._log_import(import_from.module)
        analyser = self._analyse_module_to_import(import_from, import_from.module)
        if analyser is not None:
            import_alias: dict[str] = \
                {alias.name: alias.asname if alias.asname is not None else alias.name for alias in import_from.names}

            new_symbols: dict[str, ISymbol] = analyser.export_symbols(list(import_alias.keys()))

            # check if the wildcard is used and filter the symbols
            if constants.IMPORT_WILDCARD in import_alias:
                import_alias.pop(constants.IMPORT_WILDCARD)
                for imported_symbol_id in new_symbols:
                    # add the symbols imported with the wildcard without specific aliases in the dict
                    if imported_symbol_id not in import_alias:
                        import_alias[imported_symbol_id] = imported_symbol_id

            # includes the module to be able to generate the functions
            imported_module = self._build_import(analyser.path, analyser.tree, analyser, import_alias)
            self._current_scope.include_symbol(import_from.module, imported_module)

            for name, alias in import_alias.items():
                if name in new_symbols:
                    self._current_scope.include_symbol(alias, imported_module.symbols[name])
                else:
                    # if there's a symbol that couldn't be loaded, log a compiler error
                    self._log_unresolved_import(import_from, name)

    def visit_Import(self, import_node: ast.Import):
        """
        Includes methods and variables from other modules into the current scope

        :param import_node:
        """
        import_alias: dict[str] = \
            {alias.name: alias.asname if alias.asname is not None else alias.name for alias in import_node.names}

        for target, alias in import_alias.items():
            self._log_import(target)
            analyser = self._analyse_module_to_import(import_node, target)
            if analyser is not None:
                new_symbols: dict[str, ISymbol] = analyser.export_symbols()
                for symbol in [symbol for symbol in analyser.symbols if symbol not in new_symbols]:
                    # if there's a symbol that couldn't be loaded, log a compiler error
                    self._log_unresolved_import(import_node, '{0}.{1}'.format(target, symbol))

                imported_module = self._build_import(analyser.path, analyser.tree, analyser)
                self._current_scope.include_symbol(alias, imported_module)

    def _build_import(self, origin: str, syntax_tree: ast.AST,
                      import_analyser: ImportAnalyser,
                      imported_symbols: dict[str, ISymbol] = None) -> Import:

        if import_analyser.is_builtin_import:
            return BuiltinImport(origin, syntax_tree, import_analyser, imported_symbols)

        return Import(origin, syntax_tree, import_analyser, imported_symbols)

    def _analyse_module_to_import(self, origin_node: ast.AST, target: str) -> ImportAnalyser | None:
        already_imported = {imported.origin: imported.analyser
                            for imported in self._current_module.symbols.values()
                            if isinstance(imported, Import) and imported.analyser is not None
                            }
        already_imported.update(self._analysed_files)

        try:
            analyser = ImportAnalyser(import_target=target,
                                      root_folder=self.root_folder,
                                      importer_file=self.filename,
                                      already_imported_modules=already_imported,
                                      import_stack=self._import_stack.copy(),
                                      log=self._log,
                                      fail_fast=self._fail_fast)

            if analyser.recursive_import:
                self._log_error(
                    CompilerError.CircularImport(line=origin_node.lineno,
                                                 col=origin_node.col_offset,
                                                 target_import=target,
                                                 target_origin=self.filename)
                )

            elif not analyser.can_be_imported:
                circular_import_error = next((error for error in analyser.errors
                                              if isinstance(error, CompilerError.CircularImport)),
                                             None)

                if circular_import_error is not None:
                    # if the problem was a circular import, the error was already logged
                    self.errors.append(circular_import_error)
                elif hasattr(analyser, 'is_namespace_package') and analyser.is_namespace_package:
                    return analyser
                else:
                    self._log_unresolved_import(origin_node, target)

            else:
                if analyser.is_import_deprecated:
                    location_tip = analyser._package.new_location if analyser._package is not None else None
                    self._log_warning(
                        CompilerWarning.DeprecatedSymbol(
                            origin_node.lineno,
                            origin_node.col_offset,
                            analyser._import_identifier,
                            location_tip
                        )
                    )
                analyser.update_external_analysed_files(self._analysed_files)
                return analyser

        except CompilerError.CompilerError as error_on_import:
            if self._log:
                self.errors.append(error_on_import)
                raise error_on_import
            else:
                self._log_error(error_on_import)

    def visit_Module(self, module: ast.Module):
        """
        Visitor of the module node

        Fills module symbol table

        :param module:
        """
        mod: Module = Module()
        self._current_module = mod
        self._scope_stack.append(SymbolScope())

        global_stmts = []
        function_stmts = []
        for stmt in module.body:
            if isinstance(stmt, (ast.FunctionDef, ast.AsyncFunctionDef)):
                function_stmts.append(stmt)
            elif not (isinstance(stmt, ast.Expr) and isinstance(stmt.value, ast.Constant)):
                # don't evaluate constant expression - for example: string for documentation
                if self.visit(stmt) is not Builtin.Event:
                    global_stmts.append(stmt)

        module.body = global_stmts + function_stmts
        for var_id, var in mod.variables.items():
            # all static fields must be initialized
            if not mod.is_variable_assigned(var_id):
                self._log_error(
                    CompilerError.UnresolvedReference(
                        line=var.origin.lineno if var.origin is not None else 0,
                        col=var.origin.col_offset if var.origin is not None else 0,
                        symbol_id=var_id
                    )
                )

        for stmt in function_stmts:
            result = self.visit(stmt)
            # don't evaluate the metadata function in the following analysers
            if result is MetadataTypeSingleton:
                module.body.remove(stmt)

        self.modules['main'] = mod
        module_scope = self._scope_stack.pop()
        for symbol_id, symbol in module_scope.symbols.items():
            if symbol_id in self._global_assigned_variables:
                mod.include_symbol(symbol_id, symbol)
                mod.assign_variable(symbol_id)

        self._global_assigned_variables.clear()
        self._current_module = None

    def visit_ClassDef(self, class_node: ast.ClassDef):
        """
        Visitor of the class node

        Includes the class in the scope of its module
        """
        bases = []
        for base in class_node.bases:
            base_type_id: Any = ast.NodeVisitor.visit(self, base)
            if isinstance(base_type_id, ast.Name):
                base_type_id = base_type_id.id

            base_symbol = self.get_symbol(base_type_id)
            # TODO: change when class inheritance with builtin types is implemented #2kq1ght
            if not isinstance(base_symbol, UserClass):
                self._log_error(
                    CompilerError.NotSupportedOperation(
                        class_node.lineno, class_node.col_offset,
                        symbol_id='class inheritance with builtins'
                    )
                )
            bases.append(base_symbol)

        # TODO: change when class inheritance with multiple bases is implemented #2kq1gmc
        if len(bases) > 1:
            self._log_error(
                CompilerError.NotSupportedOperation(
                    class_node.lineno, class_node.col_offset,
                    symbol_id='class inheritance with multiple bases'
                )
            )

        # TODO: change when base classes with keyword is implemented #2kq1gqy
        if len(class_node.keywords) > 0:
            self._log_error(
                CompilerError.NotSupportedOperation(
                    class_node.lineno, class_node.col_offset,
                    symbol_id='class keyword'
                )
            )

        # TODO: change when class decorators are implemented #2ewf04r
        class_decorators: list[Method] = self._get_decorators(class_node)
        if not all(isinstance(decorator, IBuiltinDecorator) and decorator.is_class_decorator
                   for decorator in class_decorators):
            # only builtin decorator are currently accepted
            self._log_error(
                CompilerError.NotSupportedOperation(
                    class_node.lineno, class_node.col_offset,
                    symbol_id='class decorator'
                )
            )

        contract_interface_decorator = next((decorator for decorator in class_decorators
                                             if isinstance(decorator, ContractDecorator)),
                                            None)

        if contract_interface_decorator is not None:
            from boa3.internal.model.type.classes.contractinterfaceclass import ContractInterfaceClass
            user_class = ContractInterfaceClass(contract_hash=contract_interface_decorator.contract_hash,
                                                identifier=class_node.name,
                                                decorators=class_decorators,
                                                bases=bases)
        else:
            user_class = UserClass(identifier=class_node.name,
                                   decorators=class_decorators,
                                   bases=bases)

        self._current_class = user_class
        if self._current_symbol_scope is not None:
            self._current_symbol_scope.include_symbol(class_node.name, user_class)
        self._scope_stack.append(SymbolScope())

        for stmt in class_node.body:
            self.visit(stmt)

        class_scope = self._scope_stack.pop()
        self._current_module.include_class(class_node.name, user_class)
        self._current_class = None

    def visit_FunctionDef(self, function: ast.FunctionDef):
        """
        Visitor of the function node

        Includes the method in the scope of its module

        :param function:
        """
        fun_decorators: list[Method] = self._get_decorators(function)

        if any(decorator is None for decorator in fun_decorators):
            self._log_error(
                CompilerError.NotSupportedOperation(
                    function.lineno, function.col_offset,
                    symbol_id='decorator'
                )
            )

        valid_decorators: list[IDecorator] = []
        for decorator in fun_decorators:
            if isinstance(decorator, IDecorator):
                if not decorator.is_function_decorator:
                    self._log_error(
                        CompilerError.NotSupportedOperation(
                            function.lineno, function.col_offset,
                            symbol_id=f'"{decorator.identifier}" decorator with function'
                        )
                    )

                decorator.update_args(function.args, self._current_scope)
                valid_decorators.append(decorator)

            # TODO: remove when user-created decorators are implemented #86a19uwzn
            elif isinstance(decorator, Method):
                self._log_error(
                    CompilerError.NotSupportedOperation(
                        function.lineno, function.col_offset,
                        symbol_id='user-created decorators'
                    )
                )

        is_static_method = (isinstance(self._current_scope, ClassType)
                            and Builtin.StaticMethodDecorator in valid_decorators)
        is_instance_method = (isinstance(self._current_scope, ClassType)
                              and Builtin.ClassMethodDecorator not in valid_decorators
                              and not is_static_method)
        is_class_constructor = is_instance_method and function.name == constants.INIT_METHOD_ID

        external_function_name = None
        if isinstance(self._current_class, ContractInterfaceClass):
            if not is_static_method:
                self._log_error(CompilerError
                                .InvalidUsage(function.lineno, function.col_offset,
                                              "Only static methods are accepted when defining contract interfaces"
                                              ))
            else:
                display_name_decorator = next((decorator for decorator in valid_decorators
                                               if isinstance(decorator, type(Builtin.ContractMethodDisplayName))),
                                              None)
                if display_name_decorator is not None:
                    external_function_name = display_name_decorator.external_name

        if is_instance_method:
            if Builtin.InstanceMethodDecorator not in valid_decorators:
                valid_decorators.append(Builtin.InstanceMethodDecorator)

            if len(function.args.args) > 0 and function.args.args[0].annotation is None:
                # set annotation to the self method
                from boa3.internal.model import set_internal_call
                self_argument = function.args.args[0]
                self_annotation = self._current_class.identifier

                self_ast_annotation = ast.parse(self_annotation).body[0].value
                set_internal_call(self_ast_annotation)

                ast.copy_location(self_ast_annotation, self_argument)
                self_argument.annotation = self_ast_annotation

        if is_class_constructor:
            # __init__ method behave like class methods
            if Builtin.ClassMethodDecorator not in valid_decorators:
                valid_decorators.append(Builtin.ClassMethodDecorator)

        fun_args: FunctionArguments = self.visit(function.args)
        if function.returns is not None:
            fun_rtype_symbol = self.visit(function.returns)
            self._check_annotation_type(function.returns)
        else:
            fun_rtype_symbol = Type.none

        # TODO: remove when dictionary unpacking operator is implemented #2kq1hbr
        if function.args.kwarg is not None:
            self._log_error(
                CompilerError.NotSupportedOperation(
                    function.lineno, function.col_offset,
                    symbol_id='** variables'
                )
            )

        # TODO: remove when keyword-only arguments are implemented #2ewewtz
        if len(function.args.kwonlyargs) > 0:
            self._log_error(
                CompilerError.NotSupportedOperation(
                    function.lineno, function.col_offset,
                    symbol_id='keyword-only arguments'
                )
            )

        if isinstance(fun_rtype_symbol, str):
            symbol = self.get_symbol(fun_rtype_symbol, origin_node=function.returns)
            if symbol is MetadataTypeSingleton:
                self._read_metadata_object(function)
                return symbol

            fun_rtype_symbol = self.get_type(symbol)

        fun_return: IType = self.get_type(fun_rtype_symbol)

        method = Method(args=fun_args.args, defaults=function.args.defaults, return_type=fun_return,
                        vararg=fun_args.vararg,
                        origin_node=function,
                        is_public=any(isinstance(decorator, type(Builtin.Public)) for decorator in fun_decorators),
                        decorators=valid_decorators,
                        external_name=external_function_name,
                        is_init=is_class_constructor)

        # debug information
        method.file_origin = self.filename.replace(os.path.sep, constants.PATH_SEPARATOR)

        if function.name in Builtin.internal_methods:
            internal_method = Builtin.internal_methods[function.name]
            if not internal_method.is_valid_deploy_method(method):
                self._log_error(
                    CompilerError.InternalIncorrectSignature(line=function.lineno,
                                                             col=function.col_offset,
                                                             expected_method=internal_method)
                )
        if function.name == constants.DEPLOY_METHOD_ID:
            self._deploy_method = method

        self._current_method = method
        self._scope_stack.append(SymbolScope())

        # don't evaluate constant expression - for example: string for documentation
        function.body = [stmt for stmt in function.body
                         if not (isinstance(stmt, ast.Expr) and isinstance(stmt.value, ast.Constant))]

        if isinstance(self._current_class, ClassType):
            method.origin_class = self._current_class
            if self._current_class.is_interface and len(function.body) > 0:
                first_instruction = function.body[0]
                if not isinstance(first_instruction, ast.Pass):
                    self._log_warning(CompilerWarning.UnreachableCode(first_instruction.lineno,
                                                                      first_instruction.col_offset))

        for stmt in function.body:
            self.visit(stmt)

        method_scope = self._scope_stack.pop()
        global_scope_symbols = self._scope_stack[0].symbols if len(self._scope_stack) > 0 else {}

        self._set_instance_variables(method_scope)
        self._set_properties(function)

        self._current_method = None
        self.__include_callable(function.name, method)

        for var_id, var in method_scope.symbols.items():
            if isinstance(var, Variable) and var_id not in self._annotated_variables:
                method.include_variable(var_id, Variable(UndefinedType, var.origin))
            else:
                method.include_symbol(var_id, var)

        self._annotated_variables.clear()

    def _get_decorators(self, node: ast.AST) -> list[Method]:
        """
        Gets a list of the symbols used to decorate the given node

        :param node: python ast node
        :return: a list with all decorators in the node. Empty if no decorator is found.
        """
        decorators = []
        if hasattr(node, 'decorator_list'):
            for decorator in node.decorator_list:
                decorator_visit = self.visit(decorator)
                if decorator_visit is None and hasattr(decorator, 'func'):
                    decorator_visit = self.visit(decorator.func)

                symbol = self.get_symbol(decorator_visit, origin_node=decorator)
                if hasattr(symbol, 'build'):
                    symbol = symbol.build(decorator, self)
                decorators.append(symbol)

        return decorators

    def _set_instance_variables(self, scope: SymbolScope):
        if (isinstance(self._current_class, UserClass)
                and isinstance(self._current_method, Method)
                and self._current_method.is_init
                and len(self._current_method.args) > 0):

            self_id = list(self._current_method.args)[0]
            for var_id, var in scope.symbols.items():
                if var_id.startswith(self_id):
                    split_name = var_id.split(constants.ATTRIBUTE_NAME_SEPARATOR)
                    if len(split_name) > 0:
                        instance_var_id = split_name[1]
                        self._current_class.include_symbol(instance_var_id, var)
                        scope.remove_symbol(var_id)

    def _set_properties(self, function: ast.FunctionDef):
        from boa3.internal.model.builtin.decorator import PropertyDecorator
        if (isinstance(self._current_class, UserClass)
                and isinstance(self._current_method, Method)
                and any(isinstance(decorator, PropertyDecorator) for decorator in self._current_method.decorators)):
            if len(self._current_method.args) < 1 or not any('self' == arg for arg in self._current_method.args):
                self._log_error(
                    CompilerError.SelfArgumentError(function.lineno, function.col_offset)
                )
            self._current_class.include_symbol(self._current_method.origin.name, Property(self._current_method))

    def visit_arguments(self, arguments: ast.arguments) -> FunctionArguments:
        """
        Visitor of the function arguments node

        :param arguments:
        :return: a dictionary that maps each argument to its identifier
        """
        fun_args = FunctionArguments()

        for arg in arguments.args:
            var_id, var = self.visit_arg(arg)  # tuple[str, Variable]
            fun_args.add_arg(var_id, var)

        if arguments.vararg is not None:
            var_id, var = self.visit_arg(arguments.vararg)  # tuple[str, Variable]
            fun_args.set_vararg(var_id, var)

        if arguments.kwarg is not None:
            var_id, var = self.visit_arg(arguments.kwarg)  # tuple[str, Variable]
            fun_args.add_kwarg(var_id, var)

        return fun_args

    def visit_arg(self, arg: ast.arg) -> tuple[str, Variable]:
        """
        Visitor of a function argument node

        :param arg:
        :return: a tuple with the identifier and the argument
        """
        var_id = arg.arg
        var_type: IType = self.get_type(arg.annotation)

        if var_type is Type.none and isinstance(arg.annotation, ast.Name):
            var_symbol: ISymbol = self.get_symbol(arg.annotation.id, origin_node=arg)
            var_type = self.get_type(var_symbol)

        self._check_annotation_type(arg.annotation, origin_node=arg)

        return var_id, Variable(var_type)

    def visit_Return(self, ret: ast.Return):
        """
        Visitor of the function return node

        If the return is a name, verifies if the symbol is defined

        :param ret: the python ast return node
        """
        if isinstance(ret.value, ast.Name):
            symbol_id = self.visit(ret.value)
            symbol = self.get_symbol(symbol_id)
            if symbol is None:
                # the symbol doesn't exist
                self._log_error(
                    CompilerError.UnresolvedReference(ret.value.lineno, ret.value.col_offset, symbol_id)
                )

        if ret.value is not None:
            self.__set_source_origin(ret.value)

    def visit_type(self, target: ast.AST) -> IType | None:
        """
        Gets the type by its identifier

        :param target: ast node to be evaluated
        :return: the type of the value inside the node. None by default
        """
        target_type = self.visit(target)  # Type:str or IType
        if isinstance(target_type, str) and not isinstance(target, ast.Str):
            symbol = self.get_symbol(target_type)
            if symbol is None:
                # the symbol doesn't exist
                self._log_error(
                    CompilerError.UnresolvedReference(target.lineno, target.col_offset, target_type)
                )
            target_type = symbol

        if target_type is None and not isinstance(target, ast.NameConstant):
            # the value type is invalid
            return None

        if not isinstance(target_type, (ast.AST, IType, IExpression)) and isinstance(target_type, ISymbol):
            self._log_error(
                CompilerError.UnresolvedReference(
                    line=target.lineno, col=target.col_offset,
                    symbol_id=str(target_type)
                )
            )

        if isinstance(target_type, ClassType):
            init = target_type.constructor_method()
            if hasattr(init, 'build'):
                args = []
                if hasattr(target, 'args'):
                    for arg in target.args:
                        result = self.visit(arg)
                        if (isinstance(result, str) and not isinstance(arg, (ast.Str, ast.Constant))
                                and result in self._current_scope.symbols):
                            result = self.get_type(self._current_scope.symbols[result])
                        args.append(result)

                init = init.build(args)
            target_type = init.return_type if init is not None else target_type

        return self.get_type(target_type)

    def get_enumerate_type(self, var_type: IType) -> IType:
        return var_type.value_type if isinstance(var_type, SequenceType) else Type.none

    def visit_Assign(self, assign: ast.Assign):
        """
        Visitor of the variable assignment node

        Includes the variable in its scope if it's the first use

        :param assign:
        """
        # multiple assignments
        if isinstance(assign.targets[0], ast.Tuple):
            self._log_error(
                CompilerError.NotSupportedOperation(assign.lineno, assign.col_offset, 'Multiple variable assignments')
            )

        elif isinstance(assign.targets[0], ast.Subscript) and isinstance(assign.targets[0].slice, ast.Slice):
            self._log_error(
                CompilerError.NotSupportedOperation(assign.lineno, assign.col_offset, 'Assigning a value into a slice')
            )

        else:
            return self._visit_assign_value(assign)

    def _visit_assign_value(self, assign: ast.Assign | ast.AugAssign) -> IType:
        targets = assign.targets if isinstance(assign, ast.Assign) else [assign.target]
        var_type = self.visit_type(assign.value)

        if var_type is Type.none and isinstance(assign.value, ast.Name):
            symbol = self.get_symbol(assign.value.id)
            if isinstance(symbol, Event):
                var_type = Builtin.Event
                self._current_event = symbol

        return_type = var_type
        for target in targets:
            var_id = self.visit(target)
            if not isinstance(var_id, ISymbol):
                return_type = self.assign_value(var_id, var_type,
                                                source_node=assign,
                                                literal_value=asthelper.literal_eval_node(assign.value))

        return return_type

    def visit_AnnAssign(self, ann_assign: ast.AnnAssign):
        """
        Visitor of the annotated variable assignment node

        Includes the variable in its scope if it's the first use

        :param ann_assign:
        """
        var_id: str = self.visit(ann_assign.target)
        var_type: IType = self.visit_type(ann_assign.annotation)
        if var_type is Builtin.Event:
            self.visit(ann_assign.value)

        self._check_annotation_type(ann_assign.annotation, ann_assign)

        if isinstance(ann_assign.value, ast.Constant):
            annotation_type = self.get_type(ann_assign.value)
            if not var_type.is_type_of(annotation_type):
                self._log_error(
                    CompilerError.MismatchedTypes(line=ann_assign.value.lineno,
                                                  col=ann_assign.value.col_offset,
                                                  expected_type_id=var_type.identifier,
                                                  actual_type_id=annotation_type.identifier)
                )

        return self.assign_value(var_id, var_type,
                                 source_node=ann_assign,
                                 assignment=ann_assign.value is not None,
                                 literal_value=asthelper.literal_eval_node(ann_assign.value)
                                 )

    def visit_AugAssign(self, aug_assign: ast.AugAssign):
        return self._visit_assign_value(aug_assign)

    def assign_value(self,
                     var_id: str,
                     var_type: IType,
                     source_node: ast.AST,
                     assignment: bool = True,
                     literal_value: Any = None) -> IType:

        if var_type is Builtin.Event and self._current_event is not None:
            if '' in self._current_module.symbols and self._current_module.symbols[''] is self._current_event:
                self._current_scope.callables[var_id] = self._current_scope.callables.pop('')
                self._current_event.name = var_id
            else:
                self._current_scope.callables[var_id] = self._current_event
            self._current_event = None
        else:
            if isinstance(self._current_scope, UserClass):
                var = Variable(var_type, source_node)
                self.__include_class_variable(var_id, var)
            else:
                self.__include_variable(var_id, var_type,
                                        source_node=source_node,
                                        assignment=assignment,
                                        literal_value=literal_value)
        return var_type

    def visit_Global(self, global_node: ast.Global):
        """
        Visitor of the global identifier node

        :param global_node:
        """
        for var_id in global_node.names:
            symbol = self.get_symbol(var_id)
            if isinstance(symbol, Variable):
                if self._current_method is not None:
                    self._current_method.include_variable(var_id, symbol, is_global=True)
                self.__include_variable(var_id, symbol.type, source_node=global_node)

    def visit_Expr(self, expr: ast.Expr):
        """
        Visitor of the atom expression node

        :param expr:
        """
        value = self.visit(expr.value)
        if isinstance(expr.value, ast.Name):
            if (
                    # it is not a symbol of the method scope
                    (self._current_method is not None and value not in self._current_method.symbols)
                    # nor it is a symbol of the module scope
                    or (self._current_module is not None and value not in self._current_module.symbols)
                    # nor it is a symbol of the global scope
                    or value not in self.symbols
            ):
                self._log_error(
                    CompilerError.UnresolvedReference(expr.value.lineno, expr.value.col_offset, value)
                )

    def visit_Subscript(self, subscript: ast.Subscript) -> str | IType:
        """
        Verifies if it is the types in the subscription are valid

        :param subscript: the python ast subscription node
        :return: if the subscript is not a symbol, returns its type. Otherwise returns the symbol id.
        :rtype: IType or str
        """
        is_internal = hasattr(subscript, 'is_internal_call') and subscript.is_internal_call
        value = self.visit(subscript.value)
        symbol = self.get_symbol(value, is_internal=is_internal, origin_node=subscript.value) if isinstance(value, str) else value

        if isinstance(subscript.ctx, ast.Load):
            if (isinstance(symbol, (Collection, MetaType))
                    and isinstance(subscript.value, (ast.Name, ast.NameConstant, ast.Attribute))):
                # for evaluating names like list[str], dict[int, bool], etc
                values_type: Iterable[IType] = self.get_values_type(subscript.slice)
                if isinstance(symbol, Collection):
                    return symbol.build_collection(*values_type)
                else:
                    return symbol.build(*values_type)

            symbol_type = self.get_type(symbol)
            if isinstance(subscript.slice, ast.Slice):
                return symbol_type

            if isinstance(symbol, UnionType) or isinstance(symbol_type, UnionType):
                if not isinstance(symbol_type, UnionType):
                    symbol_type = symbol
                if isinstance(subscript.slice, ast.Tuple):
                    union_types = [self.get_type(value) for value in subscript.slice.elts]
                else:
                    union_types = self.get_type(subscript.slice)
                return symbol_type.build(union_types)

            if isinstance(symbol_type, Collection):
                return symbol_type.item_type

        return value

    def get_values_type(self, value: ast.AST) -> Iterable[IType | None]:
        """
        Verifies if it is a multiple assignments statement

        :param value: the python ast subscription node
        """
        value_type: IType | None = None

        if isinstance(value, (ast.Subscript, ast.Attribute, ast.Tuple, ast.BinOp)):
            # index is another subscription
            value_type = self.visit(value)
        elif isinstance(value, ast.Name):
            # index is an identifier
            value_type = self.get_symbol(value.id)

        types: Iterable[IType | None] = value_type if isinstance(value_type, Iterable) else [value_type]
        for tpe in types:
            if not isinstance(tpe, IType):
                # type hint not using identifiers or using identifiers that are not types
                index = self.visit(value)
                if isinstance(index, str):
                    self._log_error(
                        CompilerError.UnresolvedReference(value.lineno, value.col_offset, index)
                    )

        return types

    def visit_BinOp(self, bin_op: ast.BinOp) -> IType | None:
        left = self.visit(bin_op.left)
        if isinstance(left, str):
            left = self.get_symbol(left)
        elif isinstance(bin_op.left, ast.Constant) and left is None:
            left = self.get_type(left)

        right = self.visit(bin_op.right)
        if isinstance(right, str):
            right = self.get_symbol(right)
        elif isinstance(bin_op.right, ast.Constant) and right is None:
            right = self.get_type(right)

        # only validate type1 | type2, other binary operations are evaluated on TypeAnalyser
        if isinstance(bin_op.op, ast.BitOr) and isinstance(left, IType) and isinstance(right, IType):
            return left.union_type(right)

        return self.generic_visit(bin_op)

    def visit_Call(self, call: ast.Call) -> IType | None:
        """
        Visitor of a function call node

        :param call: the python ast function call node
        :return: the result type of the called function. None if the function is not found
        """
        func_id = self.visit(call.func)
        func_symbol = self.get_symbol(func_id)

        # if func_symbol is None, the called function may be a function written after in the code
        # that's why it shouldn't log a compiler error here
        if func_symbol is None:
            return None

        if not isinstance(func_symbol, Callable):
            # verify if it is a builtin method with its name shadowed
            func = Builtin.get_symbol(func_id)
            func_symbol = func if func is not None else func_symbol

            if func_symbol is Type.exception:
                func_symbol = Builtin.Exception
            elif isinstance(func_symbol, ClassType):
                func_symbol = func_symbol.constructor_method()

        if not isinstance(func_symbol, Callable):
            # the symbol doesn't exist
            self._log_error(
                CompilerError.UnresolvedReference(call.func.lineno, call.func.col_offset, func_id)
            )
        elif isinstance(func_symbol, IBuiltinMethod):
            if func_symbol.body is not None:
                self._builtin_functions_to_visit[func_id] = func_symbol
            elif func_symbol is Builtin.NewEvent:
                new_event = self.create_new_event(call)
                self.__include_callable(new_event.identifier, new_event)
                self._current_event = new_event
            else:
                args_types = [self.get_type(arg, use_metatype=True) for arg in call.args]
                updated_symbol = func_symbol.build(args_types)

                if updated_symbol.identifier != func_id:
                    self.__include_callable(updated_symbol.identifier, updated_symbol)
                    return self.get_type(updated_symbol)
        elif isinstance(func_symbol, NotifyMethod) and len(call.args) == 2:
            if isinstance(call.args[1], ast.Constant):
                notification_name = self.visit(call.args[1])
                named_notification = Event(notification_name, {'state': Variable(Type.any)})
                named_notification.add_call_origin(call)
                self._current_module.include_symbol(named_notification.identifier, named_notification)
            else:
                self._log_error(
                    CompilerError.InvalidUsage(
                        call.args[1].lineno, call.args[1].col_offset,
                        'The notification name should be a constant, otherwise it won\'t be added to the manifest')
                )

        return self.get_type(call.func)

    def create_new_event(self, create_call: ast.Call) -> Event:
        event_args = create_call.args
        args = {}
        name = Builtin.Event.identifier

        if len(event_args) < 0:
            self._log_error(
                CompilerError.UnfilledArgument(line=create_call.lineno,
                                               col=create_call.col_offset,
                                               param=list(Builtin.NewEvent.args)[0])
            )
        elif len(event_args) > 0:
            args_type = self.get_type(event_args[0])
            expected_type = Builtin.NewEvent.arguments_type
            if not Type.list.is_type_of(args_type):
                self._log_error(
                    CompilerError.MismatchedTypes(line=event_args[0].lineno,
                                                  col=event_args[0].col_offset,
                                                  expected_type_id=expected_type.identifier,
                                                  actual_type_id=args_type.identifier)
                )
            else:
                expected_type = expected_type.value_type
                for value in event_args[0].elts:
                    if not isinstance(value, ast.Tuple) or len(value.elts) < 2:
                        actual_type = self.get_type(value)
                        self._log_error(
                            CompilerError.MismatchedTypes(line=value.lineno,
                                                          col=value.col_offset,
                                                          expected_type_id=expected_type.identifier,
                                                          actual_type_id=actual_type.identifier)
                        )
                    else:
                        event_arg_name, event_arg_type = value.elts
                        are_types_valid = True
                        if not isinstance(event_arg_name, ast.Str):
                            are_types_valid = False
                        else:
                            if isinstance(event_arg_type, ast.Name):  # if is name, get the type of its id
                                arg_type = self.get_symbol(event_arg_type.id)
                            else:  # otherwise, if the result is a type
                                arg_type = self.visit(event_arg_type)

                            from boa3.internal.neo.vm.type.AbiType import AbiType
                            if not isinstance(arg_type, IType):
                                are_types_valid = False
                            elif arg_type.abi_type is AbiType.InteropInterface:
                                self._log_error(
                                    CompilerError.MismatchedTypes(line=event_arg_type.lineno,
                                                                  col=event_arg_type.col_offset,
                                                                  expected_type_id=MetaType.build().identifier,
                                                                  actual_type_id=arg_type.identifier)
                                )

                        if not are_types_valid:
                            self._log_error(
                                CompilerError.MismatchedTypes(line=value.lineno,
                                                              col=value.col_offset,
                                                              expected_type_id=expected_type.identifier,
                                                              actual_type_id=self.get_type(value).identifier)
                            )
                        else:
                            arg_name = event_arg_name.s
                            arg_type = (self.get_symbol(event_arg_type.id)
                                        if isinstance(event_arg_type, ast.Name)
                                        else self.visit(event_arg_type))
                            args[arg_name] = Variable(arg_type)

            if len(event_args) > 1:
                if not isinstance(event_args[1], ast.Str):
                    name_type = self.get_type(event_args[1])
                    self._log_error(
                        CompilerError.MismatchedTypes(line=event_args[1].lineno,
                                                      col=event_args[1].col_offset,
                                                      expected_type_id=Type.str.identifier,
                                                      actual_type_id=name_type.identifier)
                    )
                else:
                    name = event_args[1].s

        event = Event(name, args)
        event._origin_node = create_call
        return event

    def visit_Attribute(self, attribute: ast.Attribute) -> ISymbol | str:
        """
        Gets the attribute inside the ast node

        :param attribute: the python ast attribute node
        :return: returns the type of the value, the attribute symbol and its id if the attribute exists.
                 Otherwise, returns None
        """
        value_id = attribute.value.id if isinstance(attribute.value, ast.Name) else None
        value: ISymbol = self.get_symbol(value_id) if value_id is not None else self.visit(attribute.value)

        if isinstance(value, Variable):
            value = value.type
        elif isinstance(value, ScriptContainerProperty):
            value = value.type

        attribute_symbol = None
        if hasattr(value, 'symbols') and attribute.attr in value.symbols:
            attribute_symbol = value.symbols[attribute.attr]
        elif isinstance(value, Package) and attribute.attr in value.inner_packages:
            attribute_symbol = value.inner_packages[attribute.attr]
        elif Builtin.get_symbol(attribute.attr) is not None:
            attribute_symbol = Builtin.get_symbol(attribute.attr)
        elif isinstance(value, UndefinedType):
            attribute_symbol = value

        attribute_id = '{0}.{1}'.format(value_id, attribute.attr)
        if attribute_symbol is not None:
            if attribute_symbol.is_deprecated:
                self._log_warning(
                    CompilerWarning.DeprecatedSymbol(
                        attribute.lineno,
                        attribute.col_offset,
                        attribute_id,
                        attribute_symbol.new_location if hasattr(attribute_symbol, 'new_location') else None
                    )
                )
            return attribute_symbol
        else:
            return attribute_id

    def visit_For(self, for_node: ast.For):
        """
        Visitor of for statement node

        :param for_node: the python ast for node
        """
        iter_type = self.get_type(for_node.iter)
        targets = self.visit(for_node.target)
        iterator_type = iter_type.value_type if hasattr(iter_type, 'value_type') else iter_type

        if isinstance(targets, str):
            self.__include_variable(targets, iterator_type, source_node=for_node.target)
        elif isinstance(iter_type, SequenceType):
            for target in targets:
                if isinstance(target, str):
                    self.__include_variable(target, iterator_type, source_node=for_node.target)

        # continue to walk through the tree
        self.generic_visit(for_node)

    def visit_Match(self, match_node: ast.Match):
        for case in match_node.cases:
            if not (isinstance(case.pattern, (ast.MatchValue, ast.MatchSingleton, ast.MatchMapping)) or
                    isinstance(case.pattern, ast.MatchAs) and case.pattern.pattern is None and case.guard is None
            ):
                self._log_error(
                    CompilerError.NotSupportedOperation(
                        case.pattern.lineno, case.pattern.col_offset,
                        symbol_id='case pattern with guard'
                    )
                )
        return super().generic_visit(match_node)

    def visit_ListComp(self, node: ast.ListComp):
        return self._visit_comprehension(node)

    def visit_SetComp(self, node: ast.SetComp):
        return self._visit_comprehension(node)

    def visit_DictComp(self, node: ast.DictComp):
        return self._visit_comprehension(node)

    def _visit_comprehension(self, node):
        # TODO: refactor when comprehension is implemented #2ewev7w #8678dw2ak
        self._log_error(
            CompilerError.NotSupportedOperation(
                node.lineno, node.col_offset,
                symbol_id='list comprehension'
            )
        )
        self.generic_visit(node)
        return node

    def visit_Name(self, name: ast.Name) -> str:
        """
        Visitor of a name node

        :param name:
        :return: the identifier of the name
        """
        return name.id

    def visit_Starred(self, node: ast.Starred):
        # TODO: refactor when starred variables are implemented #2kq1hzg
        self._log_error(
            CompilerError.NotSupportedOperation(
                node.lineno, node.col_offset,
                symbol_id='* variables'
            )
        )
        return node.value

    def visit_Constant(self, constant: ast.Constant) -> Any:
        """
        Visitor of constant values node

        :param constant: the python ast constant value node
        :return: the value of the constant
        """
        return constant.value

    def visit_NameConstant(self, constant: ast.NameConstant) -> Any:
        """
        Visitor of constant names node

        :param constant:
        :return: the value of the constant
        """
        return constant.value

    def visit_Num(self, num: ast.Num) -> int:
        """
        Visitor of literal number node

        :param num:
        :return: the value of the number
        """
        return num.n

    def visit_Str(self, str: ast.Str) -> str:
        """
        Visitor of literal string node

        :param str:
        :return: the value of the string
        """
        return str.s

    def visit_JoinedStr(self, joined_str: ast.JoinedStr):
        """
        Visitor of joined string node

        :param joined_str:
        :return: the value of the string

        """
        for node in joined_str.values:
            self.visit(node)

    def visit_Bytes(self, btes: ast.Bytes) -> bytes:
        """
        Visitor of literal string node

        :param btes:
        :return: the value of the string
        """
        return btes.s

    def visit_Tuple(self, tup_node: ast.Tuple) -> tuple | None:
        """
        Visitor of literal tuple node

        :param tup_node: the python ast string node
        :return: the value of the tuple
        """
        result = [self.get_type(value) for value in tup_node.elts]
        if Type.none in result and isinstance(tup_node.ctx, ast.Load):
            # if can't define the type of any value, let it to be defined in the type checking
            # only in load ctx because on store ctx means a tuple of variables to be assigned
            return None
        return tuple(result)

    def visit_List(self, list_node: ast.List) -> list[Any] | None:
        """
        Visitor of literal list node

        :param list_node: the python ast list node
        :return: the value of the list
        """
        result = [self.get_type(value) for value in list_node.elts]
        if Type.none in result:
            # if can't define the type of any value, let it to be defined in the type checking
            return None
        return result

    def visit_Dict(self, dict_node: ast.Dict) -> dict[Any, Any] | None:
        """
        Visitor of literal dict node

        :param dict_node: the python ast dict node
        :return: the value of the dict
        """
        dictionary = {}
        size = min(len(dict_node.keys), len(dict_node.values))
        for index in range(size):
            key = self.get_type(dict_node.keys[index])
            value = self.get_type(dict_node.values[index])
            if key in dictionary and dictionary[key] != value:
                dictionary[key] = Type.get_generic_type(dictionary[key], value)
            else:
                dictionary[key] = value

        keys = set(dictionary.keys())
        values = set(dictionary.values())

        if Type.none in keys or Type.none in values:
            # if can't define the type of any key or value, let it to be defined in the type checking
            return None
        return dictionary

    def visit_Delete(self, node: ast.Delete) -> Any:
        """
        Visitor of delete statement node. Currently, it's not supported
        """
        self._log_error(
            CompilerError.NotSupportedOperation(
                node.lineno, node.col_offset,
                symbol_id='del keyword'
            )
        )
        self.generic_visit(node)
        return node

    # endregion



import abc
from typing import TypeVar, Self

T = TypeVar("T")


class IStack(abc.ABC):
    def __init__(self, stack_type: type[T]):
        self._stack_type = stack_type
        self._stack: list[T] = []

    @abc.abstractmethod
    def _default_constructor_args(self) -> tuple:
        return (self._stack_type,)

    def append(self, value: T):
        return self._stack.append(value)

    def clear(self):
        return self._stack.clear()

    def copy(self) -> Self:
        new_stack = self.__class__(*self._default_constructor_args())
        new_stack._stack = self._stack.copy()
        return new_stack

    def pop(self, index: int) -> T:
        return self._stack.pop(index)

    def __len__(self) -> int:
        return len(self._stack)

    def __getitem__(self, index_or_slice: int | slice):
        return self._stack[index_or_slice]

    def reverse(self, start: int = 0, end: int = None, *, rotate: bool = False):
        if end is None:
            end = len(self._stack)

        if rotate:
            reverse = self._stack[start:end]
            reverse.append(reverse.pop(0))
        else:
            reverse = list(reversed(self._stack[start:end]))
        self._stack[start:end] = reverse



from boa3.internal.model.builtin.builtin import Builtin
from boa3.internal.model.builtin.interop.iterator import IteratorType
from boa3.internal.model.standards.neostandard import INeoStandard
from boa3.internal.model.standards.standardmethod import StandardMethod
from boa3.internal.model.type.collection.sequence.uint160type import UInt160Type
from boa3.internal.model.type.type import Type


class Nep11DivisibleStandard(INeoStandard):
    def __init__(self):
        type_uint160 = UInt160Type.build()
        type_iterator = IteratorType.build()
        type_token_id = Type.union.build([Type.str, Type.bytes])

        methods = [
            StandardMethod('symbol', safe=True,
                           args={},
                           return_type=Type.str),
            StandardMethod('decimals', safe=True,
                           args={},
                           return_type=Type.int),
            StandardMethod('totalSupply', safe=True,
                           args={},
                           return_type=Type.int),
            StandardMethod('balanceOf', safe=True,
                           args={'owner': type_uint160},
                           return_type=Type.int),
            StandardMethod('tokensOf', safe=True,
                           args={'owner': type_uint160},
                           return_type=type_iterator),
            StandardMethod('transfer',
                           args={'to': type_uint160,
                                 'tokenId': type_token_id,
                                 'data': Type.any},
                           return_type=Type.bool,
                           literal_implementation=False),
            # exclusive divisible methods below
            StandardMethod('transfer',
                           args={'from': type_uint160,
                                 'to': type_uint160,
                                 'amount': Type.int,
                                 'tokenId': type_token_id,
                                 'data': Type.any},
                           return_type=Type.bool,
                           literal_implementation=False),
            StandardMethod('ownerOf', safe=True,
                           args={'tokenId': type_token_id},
                           return_type=type_iterator,
                           literal_implementation=False),
            StandardMethod('balanceOf', safe=True,
                           args={'owner': type_uint160,
                                 'tokenId': type_token_id},
                           return_type=Type.int,
                           literal_implementation=False),
        ]

        optionals = [
            StandardMethod('tokens', safe=True,
                           args={},
                           return_type=type_iterator),
            StandardMethod('properties', safe=True,
                           args={
                               'tokenId': type_token_id,
                           },
                           return_type=Type.dict,
                           literal_implementation=False),
        ]

        events = [
            Builtin.Nep11Transfer
        ]

        super().__init__(methods, events, optionals)



from boa3_test.test_drive.model.network.payloads.witnessscope import WitnessScope
from boa3_test.test_drive.model.wallet.account import Account
from boa3_test.test_drive.neoxp.command import neoexpresscommand as neoxp
from boa3_test.test_drive.neoxp.command.neoexpresscommand import NeoExpressCommand


def deploy(nef_path: str, account: Account, witness_scope: WitnessScope = None,
           force: bool = False, trace: bool = False) -> NeoExpressCommand:
    return neoxp.contract.ContractDeployCommand(nef_path, account,
                                                witness_scope=witness_scope,
                                                force=force, trace=trace)


def run(contract_id: str, method: str, *args: str, account: Account,
        witness_scope: WitnessScope = None, trace: bool = False) -> NeoExpressCommand:
    return neoxp.contract.ContractRunCommand(contract_id, method, *args,
                                             account=account,
                                             witness_scope=witness_scope,
                                             trace=trace)


def invoke(invoke_file: str, account: Account, witness_scope: WitnessScope = None,
           trace: bool = False) -> NeoExpressCommand:
    return neoxp.contract.ContractInvokeCommand(invoke_file, account,
                                                witness_scope=witness_scope,
                                                trace=trace)



from typing import Any

from boa3.internal import constants
from boa3.internal.model.type.itype import IType
from boa3.internal.model.type.primitive.primitivetype import PrimitiveType
from boa3.internal.neo.vm.type.AbiType import AbiType
from boa3.internal.neo.vm.type.StackItem import StackItemType


class IntType(PrimitiveType):
    """
    A class used to represent Python int type
    """

    def __init__(self):
        identifier = 'int'
        super().__init__(identifier)

    @property
    def default_value(self) -> Any:
        return int()

    @property
    def abi_type(self) -> AbiType:
        return AbiType.Integer

    @property
    def stack_item(self) -> StackItemType:
        return StackItemType.Integer

    def _init_class_symbols(self):
        super()._init_class_symbols()

        from boa3.internal.model.builtin.builtin import Builtin

        self._instance_methods[constants.INIT_METHOD_ID] = Builtin.IntInt

    @classmethod
    def build(cls, value: Any) -> IType:
        if cls._is_type_of(value):
            from boa3.internal.model.type.type import Type
            return Type.int

    @classmethod
    def _is_type_of(cls, value: Any):
        return type(value) is int or isinstance(value, IntType)



import ast

from boa3.internal.model.builtin.contract.nep17_interface_methods.nep17interfacemethod import Nep17InterfaceMethod
from boa3.internal.model.builtin.interop.contract import ContractType
from boa3.internal.model.variable import Variable


class Nep17InterfaceTransferMethod(Nep17InterfaceMethod):

    def __init__(self, self_type: ContractType):
        from boa3.internal.model.type.type import Type
        from boa3.internal.model.type.collection.sequence.uint160type import UInt160Type

        args: dict[str, Variable] = {
            'self': Variable(self_type),
            'from_address': Variable(UInt160Type.build()),
            'to_address': Variable(UInt160Type.build()),
            'amount': Variable(Type.int),
            'data': Variable(Type.any),
        }

        data_default = ast.parse("{0}".format(Type.any.default_value)
                                 ).body[0].value

        super().__init__(args, 'transfer', return_type=Type.bool, defaults=[data_default])



from typing import Any, Self

from boa3.internal.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.internal.model.expression import IExpression
from boa3.internal.model.method import Method
from boa3.internal.model.property import Property
from boa3.internal.model.type.classes.classarraytype import ClassArrayType
from boa3.internal.model.type.collection.sequence.uint160type import UInt160Type
from boa3.internal.model.variable import Variable


class NotificationType(ClassArrayType):
    """
    A class used to represent Neo Notification class
    """

    def __init__(self):
        super().__init__('Notification')

        from boa3.internal.model.type.type import Type
        self._variables: dict[str, Variable] = {
            'script_hash': Variable(UInt160Type.build()),
            'event_name': Variable(Type.str),
            'state': Variable(Type.tuple)
        }
        self._constructor: Method = None

    @property
    def instance_variables(self) -> dict[str, Variable]:
        return self._variables.copy()

    @property
    def class_variables(self) -> dict[str, Variable]:
        return {}

    @property
    def properties(self) -> dict[str, Property]:
        return {}

    @property
    def static_methods(self) -> dict[str, Method]:
        return {}

    @property
    def class_methods(self) -> dict[str, Method]:
        return {}

    @property
    def instance_methods(self) -> dict[str, Method]:
        return {}

    def constructor_method(self) -> Method | None:
        # was having a problem with recursive import
        if self._constructor is None:
            self._constructor: Method = NotificationMethod(self)
        return self._constructor

    @classmethod
    def build(cls, value: Any = None) -> Self:
        if value is None or cls._is_type_of(value):
            return _Notification

    @classmethod
    def _is_type_of(cls, value: Any):
        return isinstance(value, NotificationType)


_Notification = NotificationType()


class NotificationMethod(IBuiltinMethod):

    def __init__(self, return_type: NotificationType):
        identifier = '-Notification__init__'
        args: dict[str, Variable] = {}
        super().__init__(identifier, args, return_type=return_type)

    def validate_parameters(self, *params: IExpression) -> bool:
        return len(params) == 0

    def generate_internal_opcodes(self, code_generator):
        from boa3.internal.neo3.core.types import UInt160

        uint160_default = UInt160.zero().to_array()

        code_generator.convert_literal(())  # state
        code_generator.convert_literal('')  # event_name
        code_generator.convert_literal(uint160_default)  # script_hash
        code_generator.convert_new_array(length=3, array_type=self.type)

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> str | None:
        return



from boa3_test.test_drive.neoxp.command.neoexpresscommand.oracle.ioraclecommand import IOracleCommand


class OracleResponseCommand(IOracleCommand):
    def __init__(self, url: str, response_path: str,
                 request_id: int | None = None,
                 trace: bool = False,
                 neo_express_data_file: str = None):

        self.request_id = request_id
        self.trace = trace
        self.input = neo_express_data_file

        command_id = 'response'
        args = [
            url,
            response_path,
        ]

        super().__init__(command_id, args)

    def _get_options(self) -> dict[str, str]:
        options = super()._get_options()

        if self.request_id:
            options[f'--request_id:{self.request_id}'] = ''

        if self.trace:
            options['--trace'] = ''

        return options



from typing import Any

from boa3.internal.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.internal.model.expression import IExpression
from boa3.internal.model.type.collection.mapping.mappingtype import MappingType
from boa3.internal.model.variable import Variable
from boa3.internal.neo.vm.opcode.Opcode import Opcode


class MapValuesMethod(IBuiltinMethod):
    def __init__(self, mapping_type: MappingType = None):
        from boa3.internal.model.type.type import Type
        if not isinstance(mapping_type, MappingType):
            self_arg = Variable(Type.mapping)
            return_type = Type.sequence
        else:
            self_arg = Variable(mapping_type)
            return_type = Type.sequence.build_collection(mapping_type.value_type)

        identifier = 'values'
        args: dict[str, Variable] = {'self': self_arg}
        super().__init__(identifier, args, return_type=return_type)

    @property
    def _arg_self(self) -> Variable:
        return self.args['self']

    def validate_parameters(self, *params: IExpression) -> bool:
        if len(params) != 1 or not isinstance(params[0], IExpression):
            return False
        return self._arg_self.type.is_type_of(params[0].type)

    def generate_internal_opcodes(self, code_generator):
        code_generator.insert_opcode(Opcode.VALUES)

    def push_self_first(self) -> bool:
        return self.has_self_argument

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> str | None:
        return None

    def build(self, value: Any) -> IBuiltinMethod:
        if type(value) == type(self.args['self'].type):
            return self
        if isinstance(value, MappingType):
            return MapValuesMethod(value)
        return super().build(value)



import abc
import ast
from typing import Self

from boa3.internal.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.internal.model.symbol import ISymbol
from boa3.internal.model.type.itype import IType
from boa3.internal.model.variable import Variable


class IInternalMethod(IBuiltinMethod, abc.ABC):

    def __init__(self,
                 identifier: str,
                 args: dict[str, Variable] = None,
                 defaults: list[ast.AST] = None,
                 return_type: IType = None,
                 vararg: tuple[str, Variable] | None = None,
                 deprecated: bool = False
                 ):
        super().__init__(identifier, args, defaults, return_type, vararg, deprecated=deprecated)

    @classmethod
    @abc.abstractmethod
    def instance(cls) -> Self:
        pass

    @classmethod
    @abc.abstractmethod
    def is_valid_deploy_method(cls, symbol: ISymbol) -> bool:
        return False



from boa3.internal.model.builtin.method.maxmethod import MaxMethod
from boa3.internal.model.type.itype import IType


class MaxByteStringMethod(MaxMethod):

    def __init__(self, arg_value: IType | None = None):
        from boa3.internal.model.type.type import Type
        is_valid_type = Type.str.is_type_of(arg_value) or Type.bytes.is_type_of(arg_value)
        super().__init__(arg_value if is_valid_type else Type.str)

    def _compare_values(self, code_generator):
        from boa3.internal.model.builtin.builtin import Builtin
        from boa3.internal.model.operation.binaryop import BinaryOp
        from boa3.internal.neo.vm.opcode.Opcode import Opcode

        code_generator.duplicate_stack_item(2)
        code_generator.duplicate_stack_item(2)

        if_is_equal = code_generator.convert_begin_if()
        code_generator.change_jump(if_is_equal, Opcode.JMPNE)
        # if str1 == str2:
        code_generator.convert_literal(True)
        #   condition = True

        if_is_equal = code_generator.convert_begin_else(if_is_equal, insert_jump=True)
        # else:
        #   limit = min((len(str1), len(str2))
        code_generator.duplicate_stack_item(2)
        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)
        code_generator.duplicate_stack_item(2)
        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)
        code_generator.convert_builtin_method_call(Builtin.Min, is_internal=True)

        #   index = 0
        code_generator.convert_literal(0)

        #   while index < limit:
        start_while = code_generator.convert_begin_while()

        #       value1 = str1[index]
        code_generator.duplicate_stack_item(4)
        code_generator.duplicate_stack_item(2)
        code_generator.convert_get_item(index_inserted_internally=True, test_is_negative_index=False)
        code_generator.duplicate_stack_item(2)

        #       value2 = str2[index]
        code_generator.duplicate_stack_item(5)
        code_generator.swap_reverse_stack_items(2)
        code_generator.convert_get_item(index_inserted_internally=True, test_is_negative_index=False)

        #       if (value2 != value1):
        code_generator.duplicate_stack_item(2)
        code_generator.duplicate_stack_item(2)
        if_values_not_equal = code_generator.convert_begin_if()
        code_generator.change_jump(if_values_not_equal, Opcode.JMPEQ)

        #           condition = value2 > value1
        code_generator.convert_operation(BinaryOp.Gt, is_internal=True)
        code_generator.remove_stack_item(2)
        code_generator.remove_stack_item(2)
        #           break
        code_generator.convert_loop_break()
        code_generator.convert_end_if(if_values_not_equal)

        #       index += 1
        code_generator.remove_stack_top_item()
        code_generator.remove_stack_top_item()
        code_generator.insert_opcode(Opcode.INC)

        #   while condition and end
        condition_address = code_generator.bytecode_size
        code_generator.duplicate_stack_item(2)
        code_generator.duplicate_stack_item(2)
        code_generator.convert_operation(BinaryOp.Gt, is_internal=True)
        while_break = code_generator.convert_end_while(start_while, condition_address)

        #   else:
        while_else = code_generator.bytecode_size
        code_generator.remove_stack_top_item()
        code_generator.remove_stack_top_item()
        #       condition = len(str1) > len(str2)
        code_generator.duplicate_stack_item(2)
        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)
        code_generator.duplicate_stack_item(2)
        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)
        code_generator.convert_operation(BinaryOp.Gt, is_internal=True)

        code_generator.convert_end_loop_else(start_while, while_else, has_else=True)

        code_generator.convert_end_if(while_break, is_internal=True)
        code_generator.convert_end_if(if_is_equal, is_internal=True)

        # if condition is True
        is_condition = code_generator.convert_begin_if()
        #   remove str2 <=> return str1
        code_generator.remove_stack_top_item()

        # else
        is_condition = code_generator.convert_begin_else(is_condition)
        #   remove str1 <=> return str2
        code_generator.remove_stack_item(2)

        code_generator.convert_end_if(is_condition)



from boa3.internal.model.builtin.builtinproperty import IBuiltinProperty
from boa3.internal.model.builtin.interop.interopmethod import InteropMethod
from boa3.internal.model.variable import Variable


class GetAddressVersionMethod(InteropMethod):
    def __init__(self):
        from boa3.internal.model.type.type import Type
        identifier = '-get_address_version'
        syscall = 'System.Runtime.GetAddressVersion'
        args: dict[str, Variable] = {}
        super().__init__(identifier, syscall, args, return_type=Type.int)


class AddressVersionProperty(IBuiltinProperty):
    def __init__(self):
        identifier = 'address_version'
        getter = GetAddressVersionMethod()
        super().__init__(identifier, getter)



from abc import ABC
from typing import Any

from boa3.internal.model.type.collection.mapping.mappingtype import MappingType
from boa3.internal.model.type.itype import IType


class MutableMappingType(MappingType, ABC):
    """
    An interface used to represent Python mutable mapping type
    """

    def __init__(self, identifier: str, keys_type: set[IType], values_type: set[IType]):
        super().__init__(identifier, keys_type, values_type)

    def is_type_of(self, value: Any) -> bool:
        if self._is_type_of(value):
            if isinstance(value, MutableMappingType):
                return self.value_type.is_type_of(value.value_type)
            return True
        return False

    @property
    def can_reassign_values(self) -> bool:
        return True



__all__ = [
    'ContractManagement',
    'CryptoLib',
    'GasToken',
    'LedgerContract',
    'NeoToken',
    'OracleContract',
    'PolicyContract',
    'RoleManagement',
    'StdLib',
]

from boa3.sc.contracts.contractmanagement import ContractManagement
from boa3.sc.contracts.cryptolib import CryptoLib
from boa3.sc.contracts.gastoken import GasToken
from boa3.sc.contracts.ledgercontract import LedgerContract
from boa3.sc.contracts.neotoken import NeoToken
from boa3.sc.contracts.oraclecontract import OracleContract
from boa3.sc.contracts.policycontract import PolicyContract
from boa3.sc.contracts.rolemanagement import RoleManagement
from boa3.sc.contracts.stdlib import StdLib



from boa3.internal.model.builtin.interop.interopmethod import InteropMethod
from boa3.internal.model.variable import Variable


class BurnGasMethod(InteropMethod):

    def __init__(self):
        from boa3.internal.model.type.type import Type
        identifier = 'burn_gas'
        syscall = 'System.Runtime.BurnGas'
        args: dict[str, Variable] = {'gas': Variable(Type.int)}
        super().__init__(identifier, syscall, args, return_type=Type.none)



from boa3.internal.model.builtin.interop.nativecontract import CryptoLibMethod
from boa3.internal.model.variable import Variable


class Sha256Method(CryptoLibMethod):
    def __init__(self):
        from boa3.internal.model.type.type import Type
        identifier = 'sha256'
        native_identifier = 'sha256'
        args: dict[str, Variable] = {'data': Variable(Type.bytes)}
        super().__init__(identifier, native_identifier, args, return_type=Type.bytes)



import ast
from typing import Any

from boa3.internal.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.internal.model.identifiedsymbol import IdentifiedSymbol
from boa3.internal.model.type.itype import IType
from boa3.internal.model.type.primitive.inttype import IntType
from boa3.internal.model.type.primitive.strtype import StrType
from boa3.internal.model.variable import Variable
from boa3.internal.neo.vm.opcode.Opcode import Opcode
from boa3.internal.model.type.type import Type


class ToBytesMethod(IBuiltinMethod):
    def __init__(self, args: dict[str, Variable] = None, defaults: list[ast.AST] = None):
        identifier = 'to_bytes'
        super().__init__(identifier, args, defaults=defaults, return_type=Type.bytes)

    @property
    def _arg_value(self) -> Variable:
        return self.args['value']

    @property
    def identifier(self) -> str:
        if isinstance(self._arg_value.type, IdentifiedSymbol):
            return '-{0}_{1}'.format(self._arg_value.type.identifier, self._identifier)
        return self._identifier

    def generate_internal_opcodes(self, code_generator):
        code_generator.convert_cast(Type.bytes, is_internal=True)

    def build(self, value: Any) -> IBuiltinMethod:
        if not isinstance(value, list):
            value = [value]

        if Type.int.is_type_of(value[0]):
            return IntToBytesMethod(value[0])
        elif Type.str.is_type_of(value[0]):
            return StrToBytesMethod(value[0])
        # if it is not a valid type, show mismatched type with int
        return IntToBytesMethod()

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> str | None:
        return None


class IntToBytesMethod(ToBytesMethod):
    def __init__(self, value_type: IType = None):
        if not isinstance(value_type, IntType):
            value_type = Type.int

        args: dict[str, Variable] = {
            'value': Variable(value_type),
            'length': Variable(Type.optional.build(Type.int)),
            'big_endian': Variable(Type.bool),
            'signed': Variable(Type.bool)
        }

        length_default = ast.parse("{0}".format(None)
                                   ).body[0].value
        big_endian_default = ast.parse("{0}".format(False)
                                       ).body[0].value
        signed_default = ast.parse("{0}".format(True)
                                   ).body[0].value

        super().__init__(args, [length_default, big_endian_default, signed_default])

    @property
    def generation_order(self) -> list[int]:
        value_index = list(self.args).index('value')
        length_index = list(self.args).index('length')
        big_endian_index = list(self.args).index('big_endian')
        signed_index = list(self.args).index('signed')

        return [big_endian_index, signed_index, length_index, value_index]

    @property
    def exception_message(self) -> str:
        return 'can not convert int to bytes'

    def generate_internal_opcodes(self, code_generator):
        from boa3.internal.model.builtin.builtin import Builtin
        from boa3.internal.model.operation.binaryop import BinaryOp
        padding_positive = b'\x00'
        padding_negative = b'\xFF'
        forgot_signed_arg_message = ', did you call to_bytes on a negative integer without setting signed=True?'
        length_arg_too_small_message = ', try raising the value of the length argument'

        # stack: big_endian, signed, length, value

        code_generator.duplicate_stack_item(2)
        code_generator.convert_literal(None)
        code_generator.convert_operation(BinaryOp.Is)
        # if length is None:
        if_len_is_null = code_generator.convert_begin_if()
        code_generator.change_jump(if_len_is_null, Opcode.JMPIFNOT)

        code_generator.swap_reverse_stack_items(2)
        code_generator.remove_stack_top_item()
        code_generator.duplicate_stack_top_item()
        super().generate_internal_opcodes(code_generator)
        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)
        #   length = len(value)
        code_generator.duplicate_stack_top_item()
        code_generator.convert_literal(0)
        #     if length == 0: length = 1
        if_len_is_null_and_value_0 = code_generator.convert_begin_if()
        code_generator.change_jump(if_len_is_null_and_value_0, Opcode.JMPNE)
        code_generator.remove_stack_top_item()
        code_generator.convert_literal(1)
        code_generator.convert_end_if(if_len_is_null_and_value_0)
        code_generator.swap_reverse_stack_items(2)
        code_generator.convert_end_if(if_len_is_null)

        code_generator.duplicate_stack_top_item()
        # if value == 0: value_bytes = b'\x00' * length
        if_int_is_zero = code_generator.convert_begin_if()
        code_generator.change_jump(if_int_is_zero, Opcode.JMPIF)
        code_generator.remove_stack_top_item()
        code_generator.remove_stack_item(2)
        code_generator.remove_stack_item(2)
        code_generator.convert_literal(padding_positive)
        code_generator.swap_reverse_stack_items(2)
        code_generator.convert_operation(BinaryOp.StrBytesMul)

        # else:
        else_int_is_not_zero = code_generator.convert_begin_else(if_int_is_zero, is_internal=True)
        #   convert value to bytes (e.g., 4660 -> b'\x12\x34')
        super().generate_internal_opcodes(code_generator)

        code_generator.swap_reverse_stack_items(2)
        code_generator.swap_reverse_stack_items(3, rotate=True)
        # stack: big_endian, value_bytes, length, signed

        # if not signed:
        if_not_signed = code_generator.convert_begin_if()
        code_generator.change_jump(if_not_signed, Opcode.JMPIF)

        #   if value < 0: raise Exception
        code_generator.duplicate_stack_item(2)
        code_generator.convert_literal(0)
        if_value_is_bigger_than_length = code_generator.convert_begin_if()
        code_generator.change_jump(if_value_is_bigger_than_length, Opcode.JMPGE)
        code_generator.convert_literal(self.exception_message + forgot_signed_arg_message)
        code_generator.convert_raise_exception()
        code_generator.convert_end_if(if_value_is_bigger_than_length, is_internal=True)

        #   last_bytes_value = value_bytes[-1]
        code_generator.duplicate_stack_item(2)
        code_generator.duplicate_stack_top_item()
        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)
        code_generator.insert_opcode(Opcode.DEC)
        code_generator.convert_literal(1)
        code_generator.convert_get_substring(is_internal=True)
        code_generator.convert_literal(padding_positive)

        #   if last_bytes_value == 0x00: value_bytes = value_bytes[:-1]
        if_starts_with_0x00 = code_generator.convert_begin_if()
        code_generator.change_jump(if_starts_with_0x00, Opcode.JMPNE)
        code_generator.swap_reverse_stack_items(2)
        code_generator.convert_literal(0)
        code_generator.duplicate_stack_item(2)
        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)
        code_generator.insert_opcode(Opcode.DEC)
        code_generator.convert_get_substring(is_internal=True)
        code_generator.swap_reverse_stack_items(2)

        code_generator.convert_end_if(if_starts_with_0x00)
        # stack: big_endian, value_bytes, length
        # else if signed:
        else_if_is_signed = code_generator.convert_begin_else(if_not_signed)

        #   if value < 0: padding = b'\xFF'
        code_generator.duplicate_stack_item(2)
        code_generator.convert_literal(0)
        if_value_negative = code_generator.convert_begin_if()
        code_generator.change_jump(if_value_negative, Opcode.JMPGE)
        code_generator.convert_literal(padding_negative)
        code_generator.swap_reverse_stack_items(3, rotate=True)
        code_generator.swap_reverse_stack_items(3, rotate=True)

        #   else if value >= 0:
        else_is_positive = code_generator.convert_begin_else(if_value_negative, is_internal=True)
        code_generator.duplicate_stack_item(2)
        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)
        code_generator.duplicate_stack_item(2)
        code_generator.convert_operation(BinaryOp.Gt)

        #       if len(value_bytes) > length: raise Exception
        if_positive_value_is_too_big = code_generator.convert_begin_if()
        code_generator.convert_literal(self.exception_message + length_arg_too_small_message)
        code_generator.convert_raise_exception()
        code_generator.convert_end_if(if_positive_value_is_too_big, is_internal=True)

        # stack: big_endian, value_bytes, length
        code_generator.convert_end_if(else_if_is_signed, is_internal=True)
        # padding = b'\x00' when positive
        code_generator.convert_literal(padding_positive)
        code_generator.swap_reverse_stack_items(3, rotate=True)
        code_generator.swap_reverse_stack_items(3, rotate=True)
        # stack: big_endian, padding, value_bytes, length

        code_generator.convert_end_if(else_is_positive, is_internal=True)

        # amount_of_zeros = length - len(value)
        code_generator.duplicate_stack_item(2)
        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)
        code_generator.convert_operation(BinaryOp.Sub)

        # if amount_of_zeros < 0: raise Exception
        code_generator.duplicate_stack_top_item()
        code_generator.convert_literal(0)
        if_value_is_bigger_than_length = code_generator.convert_begin_if()
        code_generator.change_jump(if_value_is_bigger_than_length, Opcode.JMPGE)
        code_generator.convert_literal(self.exception_message + length_arg_too_small_message)
        code_generator.convert_raise_exception()
        code_generator.convert_end_if(if_value_is_bigger_than_length, is_internal=True)

        # zero_bytes = amount_of_zeros * padding_positive
        code_generator.swap_reverse_stack_items(3, rotate=True)
        code_generator.swap_reverse_stack_items(2)
        code_generator.convert_operation(BinaryOp.StrBytesMul)
        code_generator.swap_reverse_stack_items(3)

        # if big_endian:
        if_is_big_endian = code_generator.convert_begin_if()
        #   result = value_bytes[::-1] (e.g., b'\x12\x34' -> b'\x34\x12')
        code_generator.convert_array_negative_stride()
        #   result = zero_bytes + result

        # else: # is little endian
        else_is_little_endian = code_generator.convert_begin_else(if_is_big_endian, is_internal=True)
        #   result = value_bytes + zero_bytes
        code_generator.swap_reverse_stack_items(2)
        code_generator.convert_end_if(else_is_little_endian, is_internal=True)

        code_generator.convert_operation(BinaryOp.Concat)

        code_generator.convert_end_if(else_int_is_not_zero, is_internal=True)


class StrToBytesMethod(ToBytesMethod):
    def __init__(self, value_type: IType = None):
        if not isinstance(value_type, StrType):
            value_type = Type.str

        args: dict[str, Variable] = {
            'value': Variable(value_type),
        }

        super().__init__(args)

    def generate_internal_opcodes(self, code_generator):
        # string and bytes' stack item are the same
        pass



from abc import ABC
from collections.abc import Iterable, Sized
from typing import Any

from boa3.internal.model.type.collection.icollection import ICollectionType
from boa3.internal.model.type.itype import IType
from boa3.internal.neo.vm.type.AbiType import AbiType
from boa3.internal.neo.vm.type.StackItem import StackItemType


class MappingType(ICollectionType, ABC):
    """
    An interface used to represent Python mapping type
    """

    def __init__(self, identifier: str, keys_type: set[IType], values_type: set[IType]):
        super().__init__(identifier, keys_type=keys_type, values_type=values_type)

    @property
    def identifier(self) -> str:
        return '{0}[{1}, {2}]'.format(self._identifier, self.key_type.identifier, self.value_type.identifier)

    @property
    def value_type(self) -> IType:
        return self.item_type

    @property
    def default_value(self) -> Any:
        return {}

    def is_type_of(self, value: Any) -> bool:
        if self._is_type_of(value):
            if isinstance(value, MappingType):
                return (self.key_type.is_type_of(value.key_type)
                        and self.value_type.is_type_of(value.value_type))
            return True
        return False

    @property
    def abi_type(self) -> AbiType:
        return AbiType.Map

    @property
    def stack_item(self) -> StackItemType:
        return StackItemType.Map

    def is_valid_key(self, key_type: IType) -> bool:
        return self.valid_key.is_type_of(key_type)

    @property
    def valid_key(self) -> IType:
        return self.key_type

    @classmethod
    def filter_types(cls, values_type) -> set[IType]:
        if values_type is None:
            values_type = set()
        elif not isinstance(values_type, set):
            if isinstance(values_type, Iterable):
                values_type = set(values_type)
            else:
                values_type = {values_type}

        if len(values_type) > 1 and all(isinstance(x, MappingType) for x in values_type):
            first_item: MappingType = list(values_type)[0]
            mapping_type = type(first_item)  # first mapping type

            k_types = set(value.key_type for value in values_type)
            v_types = set(value.value_type for value in values_type)

            if all(isinstance(x, mapping_type) for x in values_type):
                values_type = {mapping_type(keys_type=k_types, values_type=v_types)}
            else:
                from boa3.internal.model.type.type import Type
                generic_type: IType = Type.get_generic_type(*values_type)
                if isinstance(generic_type, MappingType):
                    values_type = {generic_type.build_collection(k_types, v_types)}

            return values_type

        # if any value is not a map, call the collection filter
        return super().filter_types(values_type)

    @classmethod
    def build(cls, value: Any) -> IType:
        if cls._is_type_of(value):
            # value is an instance of mapping
            if isinstance(value, dict):
                keys = list(value.keys())
                values = list(value.values())
            else:
                keys = value.key_type
                values = value.value_type

            keys_types: set[IType] = cls.get_types(keys)
            values_types: set[IType] = cls.get_types(values)
            return cls(keys_types, values_types)

        elif isinstance(value, Sized) and len(value) == 2:
            # value is a tuple with two lists of types for contructing the map
            keys_type, values_type = value

            if not isinstance(keys_type, Iterable):
                keys_type = {keys_type}
            else:
                keys_type = set(keys_type)

            if not isinstance(values_type, Iterable):
                values_types = {values_type}
            else:
                values_types = set(values_type)

            if all(isinstance(k, IType) for k in keys_type) and all(isinstance(v, IType) for v in values_types):
                return cls(keys_type, values_type)

        return super(MappingType, cls).build(value)



from typing import Any

from boa3.internal.model.builtin.interop.interopinterfacetype import InteropInterfaceType
from boa3.internal.model.method import Method
from boa3.internal.model.type.itype import IType


class Bls12381Type(InteropInterfaceType):
    """
    A class used to represent Neo BLS12_281 type
    """

    def __init__(self):
        super().__init__(identifier='IBls12381')
        self._constructor: Method = None

    @classmethod
    def build(cls, value: Any = None) -> IType:
        if value is None or cls._is_type_of(value):
            return _Bls12381

    @classmethod
    def _is_type_of(cls, value: Any):
        return isinstance(value, Bls12381Type)

    @property
    def class_variables(self):
        return {}

    @property
    def instance_variables(self):
        return {}

    @property
    def properties(self):
        return {}

    @property
    def static_methods(self):
        return {}

    @property
    def class_methods(self):
        return {}

    @property
    def instance_methods(self):
        return {}

    def constructor_method(self):
        return self._constructor


_Bls12381 = Bls12381Type()



from boa3.internal.model.builtin.interop.storage.getcheck.istoragetrygetmethod import IStorageTryGetMethod


class StorageTryGetStrMethod(IStorageTryGetMethod):
    def __init__(self):
        from boa3.internal.model.type.type import Type

        identifier = 'try_get_str'
        value_type = Type.str

        super().__init__(identifier, value_type=value_type)

    def generate_default_value_opcodes(self, code_generator):
        # default_value = ''
        code_generator.convert_literal('')

    def generate_deserialize_value_opcodes(self, code_generator):
        from boa3.internal.model.builtin.method.tostrmethod import ToStr
        from boa3.internal.model.type.type import Type

        converter = ToStr.build(Type.bytes)
        code_generator.convert_builtin_method_call(converter)



from typing import Any

from boa3.internal.model.builtin.method import IBuiltinMethod
from boa3.internal.model.builtin.method.strmethod import StrMethod
from boa3.internal.model.type.itype import IType
from boa3.internal.model.type.type import Type
from boa3.internal.model.variable import Variable


class StrSequenceMethod(StrMethod):

    def __init__(self, arg_value: IType | None = None):
        if arg_value is None:
            arg_value = Type.sequence

        args: dict[str, Variable] = {
            'object': Variable(arg_value),
        }

        super().__init__(args)

    def build(self, value: Any) -> IBuiltinMethod:
        if 'object' in self.args and self.args['object'].type is not Type.any:
            return self

        if Type.sequence.is_type_of(value):
            return StrSequenceMethod(value)

        return super().build(value)

    def generate_internal_opcodes(self, code_generator):
        from boa3.internal.model.builtin.interop.interop import Interop

        code_generator.convert_builtin_method_call(Interop.JsonSerialize, is_internal=True)



from boa3.internal.model.builtin.interop.nativecontract import CryptoLibMethod
from boa3.internal.model.variable import Variable


class Bls12381MulMethod(CryptoLibMethod):

    def __init__(self):
        from boa3.internal.model.type.type import Type
        from boa3.internal.model.builtin.interop.crypto import Bls12381Type

        bls12_381_type = Bls12381Type.build()

        identifier = 'bls12_381_mul'
        native_identifier = 'bls12381Mul'
        args: dict[str, Variable] = {
            'x': Variable(bls12_381_type),
            'mul': Variable(Type.bytes),
            'neg': Variable(Type.bool),
        }
        super().__init__(identifier, native_identifier, args, return_type=bls12_381_type)



import abc
from typing import Any

from boa3.internal.model.type.collection.sequence.sequencetype import SequenceType
from boa3.internal.model.type.itype import IType
from boa3.internal.model.type.primitive.primitivetype import PrimitiveType
from boa3.internal.neo.vm.type.AbiType import AbiType
from boa3.internal.neo.vm.type.StackItem import StackItemType


class IByteStringType(SequenceType, PrimitiveType, abc.ABC):
    """
    A class used to represent Neo ByteString type
    """

    def __init__(self, identifier: str, values_type: list[IType]):
        super().__init__(identifier, values_type)

    @property
    def identifier(self) -> str:
        return self._identifier

    @property
    def abi_type(self) -> AbiType:
        return AbiType.ByteArray

    @property
    def stack_item(self) -> StackItemType:
        return StackItemType.ByteString

    def is_type_of(self, value: Any) -> bool:
        return self._is_type_of(value)

    def _init_class_symbols(self):
        super()._init_class_symbols()

        from boa3.internal.model.builtin.builtin import Builtin

        instance_methods = [Builtin.BytesStringIsDigit,
                            Builtin.BytesStringJoin,
                            Builtin.BytesStringLower,
                            Builtin.BytesStringUpper,
                            Builtin.BytesStringStartswith,
                            Builtin.BytesStringStrip,
                            Builtin.BytesStringReplace,
                            ]

        for instance_method in instance_methods:
            self._instance_methods[instance_method.raw_identifier] = instance_method.build(self)

    def is_valid_key(self, key_type: IType) -> bool:
        return key_type == self.valid_key

    @property
    def valid_key(self) -> IType:
        from boa3.internal.model.type.type import Type
        return Type.int

    @property
    def can_reassign_values(self) -> bool:
        return False



import abc

from boa3_test.test_drive.neoxp.command.neoexpresscommand.neoexpresscommand import NeoExpressCommand


class IContractCommand(abc.ABC, NeoExpressCommand):
    def __init__(self, sub_command_id: str, args: list[str] = None):
        command_id = 'contract'
        if len(sub_command_id) > 0:
            command_id = f'{command_id} {sub_command_id}'

        super().__init__(command_id, args)



from enum import IntEnum


class WitnessScope(IntEnum):
    """
    Determine the rules for a smart contract :func:`CheckWitness()` sys call.
    """
    #: No Contract was witnessed. Only sign the transaction.
    NONE = 0x0
    #: Allow the witness if the current calling script hash equals the entry script hash into the virtual machine.
    #: Using this prevents passing :func:`CheckWitness()` in a smart contract called via another smart contract.
    CALLED_BY_ENTRY = 0x01
    #: Allow the witness if called from a smart contract that is whitelisted in the signer
    #: :attr:`~neo3.network.payloads.verification.Signer.allowed_contracts` attribute.
    CUSTOM_CONTRACTS = 0x10
    #: Allow the witness if any public key is in the signer
    #: :attr:`~neo3.network.payloads.verification.Signer.allowed_groups` attribute is whitelisted in the contracts
    #: manifest.groups array.
    CUSTOM_GROUPS = 0x20
    #: Allow the witness if the specified :attr:`~neo3.network.payloads.verification.Signer.rules` are satisfied
    WITNESS_RULES = 0x40
    #: Allow the witness in all context. Equal to NEO 2.x's default behaviour.
    GLOBAL = 0x80


class WitnessRuleAction(IntEnum):
    DENY = 0
    ALLOW = 1


class WitnessConditionType(IntEnum):
    BOOLEAN = 0x0
    NOT = 0x01
    AND = 0x2
    OR = 0x03
    SCRIPT_HASH = 0x18
    GROUP = 0x19
    CALLED_BY_ENTRY = 0x20
    CALLED_BY_CONTRACT = 0x28
    CALLED_BY_GROUP = 0x29



__all__ = ['FindOptionsType',
           'StorageContextType',
           'StorageDeleteMethod',
           'StorageFindMethod',
           'StorageGetContextMethod',
           'StorageGetReadOnlyContextMethod',
           'StorageGetBoolMethod',
           'StorageGetBytesMethod',
           'StorageGetDictMethod',
           'StorageGetECPointMethod',
           'StorageGetIntMethod',
           'StorageGetListMethod',
           'StorageGetObjectMethod',
           'StorageGetStrMethod',
           'StorageGetUInt160Method',
           'StorageGetUInt256Method',
           'StorageMapType',
           'StoragePutBoolMethod',
           'StoragePutBytesMethod',
           'StoragePutDictMethod',
           'StoragePutECPointMethod',
           'StoragePutIntMethod',
           'StoragePutListMethod',
           'StoragePutObjectMethod',
           'StoragePutStrMethod',
           'StoragePutUInt160Method',
           'StoragePutUInt256Method',
           'StorageTryGetBoolMethod',
           'StorageTryGetBytesMethod',
           'StorageTryGetDictMethod',
           'StorageTryGetECPointMethod',
           'StorageTryGetIntMethod',
           'StorageTryGetListMethod',
           'StorageTryGetObjectMethod',
           'StorageTryGetStrMethod',
           'StorageTryGetUInt160Method',
           'StorageTryGetUInt256Method',
           ]

from boa3.internal.model.builtin.interop.storage.findoptionstype import FindOptionsType
from boa3.internal.model.builtin.interop.storage.get import *
from boa3.internal.model.builtin.interop.storage.getcheck import *
from boa3.internal.model.builtin.interop.storage.put import *
from boa3.internal.model.builtin.interop.storage.storagecontext import StorageContextType
from boa3.internal.model.builtin.interop.storage.storagedeletemethod import StorageDeleteMethod
from boa3.internal.model.builtin.interop.storage.storagefindmethod import StorageFindMethod
from boa3.internal.model.builtin.interop.storage.storagegetcontextmethod import StorageGetContextMethod
from boa3.internal.model.builtin.interop.storage.storagegetreadonlycontextmethod import StorageGetReadOnlyContextMethod
from boa3.internal.model.builtin.interop.storage.storagemap.storagemaptype import StorageMapType



from boa3.internal.model.operation.operator import Operator
from boa3.internal.model.operation.unary.unaryoperation import UnaryOperation
from boa3.internal.model.type.enum.intflagtype import IntFlagType
from boa3.internal.model.type.type import IType, Type
from boa3.internal.neo.vm.opcode.Opcode import Opcode


class LogicNot(UnaryOperation):
    """
    A class used to represent the bit not [ ~ ] operation

    :ivar operator: the operator of the operation. Inherited from :class:`IOperation`
    :ivar operand: the operand type. Inherited from :class:`UnaryOperation`
    :ivar result: the result type of the operation.  Inherited from :class:`IOperation`
    """
    _valid_types: list[IType] = [Type.int, Type.bool]

    def __init__(self, operand: IType = Type.int):
        self.operator: Operator = Operator.BitNot
        super().__init__(operand)

    def validate_type(self, *types: IType) -> bool:
        if len(types) != self.number_of_operands:
            return False
        operand: IType = types[0]

        return any(_type.is_type_of(operand) for _type in self._valid_types)

    def _get_result(self, operand: IType) -> IType:
        if self.validate_type(operand):
            if isinstance(operand, IntFlagType):
                return operand
            return Type.int

        return Type.none

    def generate_opcodes(self, code_generator):
        from boa3.internal.model.operation.binaryop import BinaryOp

        if isinstance(self.operand_type, IntFlagType):
            next_power_of_two = self.operand_type.next_power_of_two

            code_generator.convert_literal(next_power_of_two - 1)
            code_generator.convert_operation(BinaryOp.Xor, is_internal=True)

        else:
            super().generate_opcodes(code_generator)

    def generate_internal_opcodes(self, code_generator):
        code_generator.insert_opcode(Opcode.INVERT)



from typing import Any

from boa3.internal.model.type.enum.intflagtype import IntFlagType
from boa3.internal.model.type.itype import IType
from boa3.internal.neo3.contracts import TriggerType as Trigger


class TriggerTypeType(IntFlagType):
    """
    A class used to represent Neo interop TriggerType type
    """

    def __init__(self):
        super().__init__(Trigger)
        self._identifier = 'TriggerType'

    @property
    def default_value(self) -> Any:
        return Trigger.ALL

    @classmethod
    def build(cls, value: Any) -> IType:
        if cls._is_type_of(value):
            from boa3.internal.model.builtin.interop.interop import Interop
            return Interop.TriggerType

    @classmethod
    def _is_type_of(cls, value: Any):
        return isinstance(value, (Trigger, TriggerTypeType))



import ast
from abc import ABC

from boa3.internal.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.internal.model.expression import IExpression
from boa3.internal.model.type.itype import IType
from boa3.internal.model.variable import Variable


class InteropMethod(IBuiltinMethod, ABC):

    def __init__(self,
                 identifier: str,
                 sys_call: str,
                 args: dict[str, Variable] = None,
                 defaults: list[ast.AST] = None,
                 return_type: IType = None,
                 deprecated: bool = False
                 ):
        self._sys_call: str = sys_call
        super().__init__(identifier, args, defaults, return_type, deprecated=deprecated)

    def validate_parameters(self, *params: IExpression) -> bool:
        if any(not isinstance(param, IExpression) for param in params):
            return False

        args: list[IType] = [arg.type for arg in self.args.values()]
        if len(params) != len(args):
            return False
        return all(args[index].is_type_of(params[index]) for index in range(len(args)))

    @property
    def interop_method_hash(self) -> bytes:
        return self._method_hash(self._sys_call)

    def _method_hash(self, method_name: str) -> bytes:
        from boa3.internal.constants import SIZE_OF_INT32
        from boa3.internal.neo import cryptography
        from boa3.internal.neo.vm.type.String import String

        return cryptography.sha256(String(method_name).to_bytes())[:SIZE_OF_INT32]

    def generate_internal_opcodes(self, code_generator):
        code_generator.insert_sys_call(self.interop_method_hash)

    def _raw_bytecode(self) -> bytes:
        from boa3.internal.neo.vm.opcode.Opcode import Opcode
        return Opcode.SYSCALL + self.interop_method_hash

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> str | None:
        return None



from boa3.internal.model.builtin.interop.interopmethod import InteropMethod
from boa3.internal.model.variable import Variable


class CheckWitnessMethod(InteropMethod):

    def __init__(self):
        from boa3.internal.model.type.type import Type
        from boa3.internal.model.type.collection.sequence.ecpointtype import ECPointType
        from boa3.internal.model.type.collection.sequence.uint160type import UInt160Type

        identifier = 'check_witness'
        syscall = 'System.Runtime.CheckWitness'
        args: dict[str, Variable] = {'hash_or_pubkey': Variable(Type.union.build([ECPointType.build(),
                                                                                  UInt160Type.build()
                                                                                  ]))}
        super().__init__(identifier, syscall, args, return_type=Type.bool)



import ast

from boa3.internal.model import set_internal_call
from boa3.internal.model.builtin.classmethod.countmethod import CountMethod
from boa3.internal.model.variable import Variable


class CountStrMethod(CountMethod):
    def __init__(self):
        from boa3.internal.model.type.type import Type

        args: dict[str, Variable] = {
            'self': Variable(Type.str),
            'value': Variable(Type.str),
            'start': Variable(Type.int),
            'end': Variable(Type.union.build([Type.int, Type.none])),
        }

        start_default = set_internal_call(ast.parse("{0}".format(Type.int.default_value)
                                                    ).body[0].value)
        end_default = set_internal_call(ast.parse("{0}".format(Type.none.default_value)
                                                  ).body[0].value)

        super().__init__(args, [start_default, end_default])

    def generate_internal_opcodes(self, code_generator):
        from boa3.internal.model.builtin.builtin import Builtin
        from boa3.internal.model.operation.binaryop import BinaryOp
        from boa3.internal.neo.vm.opcode.Opcode import Opcode

        # if end is None
        code_generator.swap_reverse_stack_items(4)
        code_generator.duplicate_stack_top_item()
        code_generator.convert_operation(BinaryOp.IsNone, is_internal=True)

        verify_end_none = code_generator.convert_begin_if()

        #   end = len(self)
        code_generator.remove_stack_top_item()
        code_generator.duplicate_stack_item(3)
        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)

        code_generator.convert_end_if(verify_end_none)

        # if end < 0:
        code_generator.duplicate_stack_top_item()
        code_generator.convert_literal(0)

        verify_end_neg = code_generator.convert_begin_if()
        code_generator.change_jump(verify_end_neg, Opcode.JMPGE)

        #   end += len(self)
        code_generator.duplicate_stack_item(4)
        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)
        code_generator.convert_operation(BinaryOp.Add, is_internal=True)

        #   if end < 0:
        code_generator.duplicate_stack_top_item()
        code_generator.convert_literal(0)

        verify_end_still_neg = code_generator.convert_begin_if()
        code_generator.change_jump(verify_end_still_neg, Opcode.JMPGE)

        #       end = 0
        code_generator.remove_stack_top_item()
        code_generator.convert_literal(0)

        code_generator.convert_end_if(verify_end_still_neg, is_internal=True)

        # else:
        verify_end_neg = code_generator.convert_begin_else(verify_end_neg, is_internal=True)

        #   if end > len(self)
        code_generator.duplicate_stack_top_item()
        code_generator.duplicate_stack_item(5)
        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)

        verify_gt_size = code_generator.convert_begin_if()
        code_generator.change_jump(verify_gt_size, Opcode.JMPLE)

        #       end = len(self)
        code_generator.remove_stack_top_item()
        code_generator.duplicate_stack_item(3)
        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)

        code_generator.convert_end_if(verify_gt_size)
        code_generator.convert_end_if(verify_end_neg)

        # if start < 0
        code_generator.duplicate_stack_item(2)
        code_generator.convert_literal(0)

        verify_start_neg = code_generator.convert_begin_if()
        code_generator.change_jump(verify_start_neg, Opcode.JMPGE)

        #   start += len(self)
        code_generator.swap_reverse_stack_items(2)
        code_generator.duplicate_stack_item(4)
        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)
        code_generator.convert_operation(BinaryOp.Add, is_internal=True)

        code_generator.duplicate_stack_top_item()
        code_generator.convert_literal(0)
        #   if start < 0:

        verify_start_still_neg = code_generator.convert_begin_if()
        code_generator.change_jump(verify_start_still_neg, Opcode.JMPGE)

        #       start = 0
        code_generator.remove_stack_top_item()
        code_generator.convert_literal(0)

        code_generator.convert_end_if(verify_start_still_neg)
        code_generator.swap_reverse_stack_items(2)

        code_generator.convert_end_if(verify_start_neg)

        # index = count = 0
        code_generator.convert_literal(0)

        # substr_size = len(substr)
        code_generator.duplicate_stack_item(4)
        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)
        code_generator.swap_reverse_stack_items(2)

        while_start = code_generator.convert_begin_while()

        # while substr_size + index <= end
        code_generator.duplicate_stack_item(6)
        code_generator.duplicate_stack_item(5)
        code_generator.duplicate_stack_item(4)
        code_generator.convert_get_substring(is_internal=True)

        #   if self[index: index + substr_size] == substr:
        code_generator.duplicate_stack_item(6)
        code_generator.convert_operation(BinaryOp.NumEq, is_internal=True)
        is_substr = code_generator.convert_begin_if()

        #       count += 1
        code_generator.insert_opcode(Opcode.INC)
        code_generator.convert_end_if(is_substr, is_internal=True)

        code_generator.swap_reverse_stack_items(4)
        #   index += 1
        code_generator.insert_opcode(Opcode.INC)
        code_generator.swap_reverse_stack_items(4)

        # while condition
        while_condition = code_generator.bytecode_size
        code_generator.duplicate_stack_item(2)
        code_generator.duplicate_stack_item(5)
        code_generator.convert_operation(BinaryOp.Add, is_internal=True)
        code_generator.duplicate_stack_item(4)
        code_generator.convert_operation(BinaryOp.LtE, is_internal=True)

        code_generator.convert_end_while(while_start, while_condition, is_internal=True)

        # clean stack
        code_generator.swap_reverse_stack_items(4)
        code_generator.remove_stack_top_item()
        code_generator.remove_stack_top_item()
        code_generator.remove_stack_top_item()
        code_generator.swap_reverse_stack_items(3)
        code_generator.remove_stack_top_item()
        code_generator.remove_stack_top_item()



import ast
from abc import ABC

from boa3.internal.model.builtin.method.builtinevent import IBuiltinEvent
from boa3.internal.model.variable import Variable


class InteropEvent(IBuiltinEvent, ABC):

    def __init__(self,
                 identifier: str,
                 sys_call: str,
                 args: dict[str, Variable] = None,
                 defaults: list[ast.AST] = None,
                 deprecated: bool = False
                 ):
        self._sys_call: str = sys_call
        super().__init__(identifier, args, defaults, deprecated=deprecated)

    @property
    def interop_method_hash(self) -> bytes:
        return self._method_hash(self._sys_call)

    def _method_hash(self, method_name: str) -> bytes:
        from boa3.internal.constants import SIZE_OF_INT32
        from boa3.internal.neo import cryptography
        from boa3.internal.neo.vm.type.String import String

        return cryptography.sha256(String(method_name).to_bytes())[:SIZE_OF_INT32]

    def generate_internal_opcodes(self, code_generator):
        code_generator.insert_sys_call(self.interop_method_hash)



from boa3.internal.model.builtin.interop.nativecontract import NeoContractMethod
from boa3.internal.model.variable import Variable


class GetCommitteeAddressMethod(NeoContractMethod):

    def __init__(self):
        from boa3.internal.model.type.collection.sequence.uint160type import UInt160Type

        identifier = 'get_committee_address'
        native_identifier = 'getCommitteeAddress'
        args: dict[str, Variable] = {}
        super().__init__(identifier, native_identifier, args, return_type=UInt160Type.build())



__all__ = [
    'OptimizationLevel',
    'is_storing_static_variable'
]

import enum

from boa3.internal.model.variable import Variable


class OptimizationLevel(enum.IntEnum):
    NONE = enum.auto()
    DEBUG = enum.auto()
    HIGH = enum.auto()

    DEFAULT = HIGH


def is_storing_static_variable(level: OptimizationLevel, var: Variable) -> bool:
    if not hasattr(var, '_set_as_global'):
        var._set_as_global = not var.is_reassigned and var.has_literal_value

    if level <= OptimizationLevel.DEBUG:
        return True

    # doesn't store variables that we know the value on higher optimization levels
    return var.is_reassigned or not var.has_literal_value



import ast
from typing import Any

from boa3.internal.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.internal.model.type.primitive.ibytestringtype import IByteStringType
from boa3.internal.model.variable import Variable
from boa3.internal.neo.vm.opcode.Opcode import Opcode


class StripMethod(IBuiltinMethod):
    def __init__(self, self_type: IByteStringType = None):
        from boa3.internal.model.type.type import Type

        if not isinstance(self_type, IByteStringType):
            self_type = Type.bytes

        identifier = 'strip'
        args: dict[str, Variable] = {
            'self': Variable(self_type),
            'chars': Variable(self_type),
        }

        from string import whitespace

        # whitespace is ' \t\n\r\v\f', but it needs to be r' \t\n\r\v\f'
        # encode('unicode-escape') will make it a bytes equivalent of r' \t\n\r\v\f' and
        # decode() will convert it back to str
        whitespace_chars = [char.encode('unicode-escape').decode() for char in whitespace]

        if Type.str.is_type_of(self_type):
            chars_default = ast.parse("'{0}'".format(''.join(whitespace_chars))).body[0].value
        else:
            chars_default = ast.parse("b'{0}'".format(''.join(whitespace_chars))).body[0].value

        super().__init__(identifier, args, defaults=[chars_default], return_type=self_type)

    @property
    def _arg_self(self) -> Variable:
        return self.args['self']

    def generate_internal_opcodes(self, code_generator):
        from boa3.internal.model.builtin.builtin import Builtin
        from boa3.internal.model.operation.binaryop import BinaryOp

        # Initialize first main loop to check the leading characters
        code_generator.duplicate_stack_item(2)
        # string_size = len(string)
        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)
        code_generator.duplicate_stack_item(2)
        # chars_size = len(chars)
        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)
        # index = 0
        code_generator.convert_literal(0)
        # stop_loop = False
        code_generator.convert_literal(False)

        # while index < string_size and not stop_loop:
        while_checking_all_leading_chars = code_generator.convert_begin_while()
        code_generator.duplicate_stack_item(5)
        code_generator.duplicate_stack_item(2)
        code_generator.convert_literal(1)
        #   char = string[index]
        code_generator.convert_get_substring(is_internal=True)
        #   index_chars = 0
        code_generator.convert_literal(0)

        #   while index < string_size:
        while_checking_leading_char = code_generator.convert_begin_while()
        code_generator.duplicate_stack_item(6)
        code_generator.duplicate_stack_item(2)
        code_generator.convert_literal(1)
        code_generator.convert_get_substring(is_internal=True)
        code_generator.duplicate_stack_item(3)
        #       equal = char == chars[index_chars]
        code_generator.convert_operation(BinaryOp.NumEq, is_internal=True)
        code_generator.swap_reverse_stack_items(2)
        #       index_chars += 1
        code_generator.insert_opcode(Opcode.INC)
        code_generator.swap_reverse_stack_items(2)
        #       if equal:
        if_leading_char_is_ws = code_generator.convert_begin_if()
        code_generator.remove_stack_top_item()
        code_generator.remove_stack_top_item()
        #           index += 1
        code_generator.insert_opcode(Opcode.INC)
        code_generator.convert_literal(False)
        #           stop_loop = False
        #           break
        code_generator.convert_loop_break()
        code_generator.convert_end_if(if_leading_char_is_ws)

        while_condition_leading_chars = code_generator.bytecode_size
        code_generator.duplicate_stack_top_item()
        code_generator.duplicate_stack_item(5)
        code_generator.convert_operation(BinaryOp.Lt, is_internal=True)
        code_generator.duplicate_stack_top_item()
        #       if index >= string_size:
        if_leading_char_is_not_ws = code_generator.convert_begin_if()
        code_generator.change_jump(if_leading_char_is_not_ws, Opcode.JMPIF)
        code_generator.remove_stack_top_item()
        code_generator.remove_stack_top_item()
        code_generator.remove_stack_top_item()
        #           stop_loop = True
        code_generator.convert_literal(True)
        #           break
        code_generator.convert_loop_break()
        code_generator.convert_end_if(if_leading_char_is_not_ws)
        code_generator.convert_end_while(while_checking_leading_char, while_condition_leading_chars, is_internal=True)

        while_condition_all_leading_chars = code_generator.bytecode_size
        #   if stop_loop:
        if_stop_while = code_generator.convert_begin_if()
        #       break
        code_generator.convert_loop_break()
        code_generator.convert_end_if(if_stop_while, is_internal=True)
        code_generator.duplicate_stack_top_item()
        code_generator.duplicate_stack_item(4)
        code_generator.convert_operation(BinaryOp.Lt, is_internal=True)
        code_generator.convert_end_while(while_checking_all_leading_chars, while_condition_all_leading_chars, is_internal=True)

        # Initialize second main loop to check the leading characters
        # n_leading = index
        code_generator.swap_reverse_stack_items(3)
        # index = string_size - 1
        code_generator.insert_opcode(Opcode.DEC)
        # stop_loop = False
        code_generator.convert_literal(False)

        # while index > string_size and not stop_loop
        while_checking_all_trailing_chars = code_generator.convert_begin_while()
        code_generator.duplicate_stack_item(5)
        code_generator.duplicate_stack_item(2)
        code_generator.convert_literal(1)
        #   char = string[index]
        code_generator.convert_get_substring(is_internal=True)
        #   index_chars = 0
        code_generator.convert_literal(0)

        #   while index < string_size:
        while_checking_trailing_char = code_generator.convert_begin_while()
        code_generator.duplicate_stack_item(6)
        code_generator.duplicate_stack_item(2)
        code_generator.convert_literal(1)
        code_generator.convert_get_substring(is_internal=True)
        code_generator.duplicate_stack_item(3)
        #       equal = char == chars[index_chars]
        code_generator.convert_operation(BinaryOp.NumEq, is_internal=True)
        code_generator.swap_reverse_stack_items(2)
        #       index_chars += 1
        code_generator.insert_opcode(Opcode.INC)
        code_generator.swap_reverse_stack_items(2)
        #       if equal:
        if_trailing_char_is_ws = code_generator.convert_begin_if()
        code_generator.remove_stack_top_item()
        code_generator.remove_stack_top_item()
        #           index -= 1
        code_generator.insert_opcode(Opcode.DEC)
        code_generator.convert_literal(False)
        #           stop_loop = False
        #           break
        code_generator.convert_loop_break()
        code_generator.convert_end_if(if_trailing_char_is_ws)

        while_condition_trailing_chars = code_generator.bytecode_size
        code_generator.duplicate_stack_top_item()
        code_generator.duplicate_stack_item(5)
        code_generator.convert_operation(BinaryOp.Lt, is_internal=True)
        code_generator.duplicate_stack_top_item()
        #       if index >= string_size:
        if_trailing_char_is_not_ws = code_generator.convert_begin_if()
        code_generator.change_jump(if_trailing_char_is_not_ws, Opcode.JMPIF)
        # remove_extras
        code_generator.remove_stack_top_item()
        code_generator.remove_stack_top_item()
        code_generator.remove_stack_top_item()
        #           stop_loop = True
        code_generator.convert_literal(True)
        #           break
        code_generator.convert_loop_break()
        code_generator.convert_end_if(if_trailing_char_is_not_ws)
        code_generator.convert_end_while(while_checking_trailing_char, while_condition_trailing_chars, is_internal=True)

        while_condition_all_trailing_chars = code_generator.bytecode_size
        #   if stop_loop:
        if_stop_while = code_generator.convert_begin_if()
        #       break
        code_generator.convert_loop_break()
        code_generator.convert_end_if(if_stop_while, is_internal=True)
        code_generator.duplicate_stack_top_item()
        code_generator.duplicate_stack_item(4)
        code_generator.convert_operation(BinaryOp.Gt, is_internal=True)
        code_generator.convert_end_while(while_checking_all_trailing_chars, while_condition_all_trailing_chars, is_internal=True)

        # Strips the string using the indexes found
        code_generator.remove_stack_item(2)
        code_generator.duplicate_stack_item(2)
        code_generator.convert_operation(BinaryOp.Sub, is_internal=True)
        code_generator.insert_opcode(Opcode.INC)
        code_generator.swap_reverse_stack_items(3)
        code_generator.remove_stack_top_item()
        code_generator.swap_reverse_stack_items(2)
        # string[n_leading:index]
        code_generator.convert_get_substring(is_internal=True)

    def push_self_first(self) -> bool:
        return self.has_self_argument

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> str | None:
        return None

    def build(self, value: Any) -> IBuiltinMethod:
        if type(value) == type(self.args['self'].type):
            return self
        if isinstance(value, IByteStringType):
            return StripMethod(value)
        return super().build(value)



from boa3.internal.model.operation.binary.binaryoperation import BinaryOperation
from boa3.internal.model.operation.operator import Operator
from boa3.internal.model.type.type import IType, Type
from boa3.internal.neo.vm.opcode.Opcode import Opcode


class Modulo(BinaryOperation):
    """
    A class used to represent a numeric modulo operation

    :ivar operator: the operator of the operation. Inherited from :class:`IOperation`
    :ivar left: the left operand type. Inherited from :class:`BinaryOperation`
    :ivar right: the left operand type. Inherited from :class:`BinaryOperation`
    :ivar result: the result type of the operation.  Inherited from :class:`IOperation`
    """
    _valid_types: list[IType] = [Type.int]

    def __init__(self, left: IType = Type.int, right: IType = None):
        self.operator: Operator = Operator.Mod
        super().__init__(left, right)

    def validate_type(self, *types: IType) -> bool:
        if len(types) != self.number_of_operands:
            return False
        left: IType = types[0]
        right: IType = types[1]

        return left == right and any(_type.is_type_of(left) for _type in self._valid_types)

    def _get_result(self, left: IType, right: IType) -> IType:
        if self.validate_type(left, right):
            return left
        else:
            return Type.none

    def generate_opcodes(self, code_generator):
        from boa3.internal.model.operation.binaryop import BinaryOp

        # neo's mod has a different result from python's mod in some cases
        code_generator.swap_reverse_stack_items(2)
        code_generator.duplicate_stack_item(2)

        # result = first % second
        super().generate_opcodes(code_generator)

        # if the result is not zero and the sign is different from the second operator the result is different
        code_generator.duplicate_stack_top_item()
        code_generator.insert_opcode(Opcode.SIGN)   # sign of result
        code_generator.duplicate_stack_item(3)
        code_generator.insert_opcode(Opcode.SIGN)   # sign of second
        code_generator.duplicate_stack_item(2)
        code_generator.convert_operation(BinaryOp.NumNotEq, is_internal=True)

        code_generator.swap_reverse_stack_items(2)
        code_generator.convert_literal(0)
        code_generator.convert_operation(BinaryOp.NumNotEq, is_internal=True)

        # if sign(result) != sign(second) and result != 0:
        code_generator.convert_operation(BinaryOp.And, is_internal=True)
        if_negative = code_generator.convert_begin_if()
        #   result += second
        code_generator.convert_operation(BinaryOp.Add, is_internal=True)
        # else
        else_negative = code_generator.convert_begin_else(if_negative, is_internal=True)
        #   # just clear stack
        code_generator.remove_stack_item(2)
        code_generator.convert_end_if(else_negative, is_internal=True)

    def generate_internal_opcodes(self, code_generator):
        code_generator.insert_opcode(Opcode.MOD)



from boa3.internal.model.builtin.interop.nativecontract import StdLibMethod
from boa3.internal.model.variable import Variable


class JsonDeserializeMethod(StdLibMethod):
    def __init__(self):
        from boa3.internal.model.type.type import Type
        identifier = 'json_deserialize'
        native_identifier = 'jsonDeserialize'
        args: dict[str, Variable] = {'json': Variable(Type.str)}
        super().__init__(identifier, native_identifier, args, return_type=Type.any)



__all__ = [
    'CodeOptimizer'
]

from boa3.internal.compiler.codegenerator.engine.executionscript import ExecutionScript
from boa3.internal.compiler.codegenerator.engine.neoengine import NeoEngine
from boa3.internal.compiler.codegenerator.optimizerhelper import OptimizationLevel
from boa3.internal.compiler.codegenerator.vmcodemapping import VMCodeMapping
from boa3.internal.model.method import Method
from boa3.internal.model.symbol import ISymbol


class CodeOptimizer:
    def __init__(self, symbol_table: dict[str, ISymbol]):
        if isinstance(symbol_table, dict):
            # works with a copy to prevent changes on the original symbol table
            symbol_table = symbol_table.copy()
        else:
            symbol_table = {}

        self.symbol_table = symbol_table
        self._map_instance = VMCodeMapping.instance()

    def optimize(self, optimization_level: OptimizationLevel = OptimizationLevel.DEFAULT):
        engine = NeoEngine()
        engine.load_script(ExecutionScript.from_code_map(self._map_instance))

        executed_instructions_addresses = set()

        # executed the methods to check if there are opcodes that are not executed
        for method in self.symbol_table.values():
            if isinstance(method, Method) and method.is_public:
                executed_instructions = engine.execute(method.start_address)

                executed_instructions_addresses.update(executed_instructions)



import ast
from abc import ABC

from boa3.internal.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.internal.model.type.itype import IType
from boa3.internal.model.variable import Variable


class ContractGetHashMethod(IBuiltinMethod, ABC):
    def __init__(self, script_hash: bytes, identifier: str, args: dict[str, Variable] = None,
                 defaults: list[ast.AST] = None, return_type: IType = None,
                 vararg: tuple[str, Variable] | None = None,
                 kwargs: dict[str, Variable] | None = None):
        self._script_hash = script_hash
        super().__init__(identifier, args, defaults, return_type, vararg, kwargs)

    @property
    def script_hash(self) -> bytes:
        return self._script_hash

    def generate_internal_opcodes(self, code_generator):
        code_generator.convert_literal(self.script_hash)



from boa3.internal.model.builtin.interop.storage.get.istoragegetmethod import IStorageGetMethod


class StorageGetBytesMethod(IStorageGetMethod):
    def __init__(self):
        from boa3.internal.model.type.type import Type

        identifier = 'get'
        value_type = Type.bytes

        super().__init__(identifier, value_type=value_type)

    def generate_default_value_opcodes(self, code_generator):
        # default_value = b''
        code_generator.convert_literal(b'')

    def generate_deserialize_value_opcodes(self, code_generator):
        # it doesn't need to deserialize bytes values
        return



from boa3_test.test_drive.neoxp.command import utils
from boa3_test.test_drive.neoxp.command.neoexpresscommand.show.ishowcommand import IShowCommand


class ShowTransactionCommand(IShowCommand):
    def __init__(self, transaction_hash: bytes | str,
                 neo_express_data_file: str = None):

        if isinstance(transaction_hash, bytes):
            transaction_hash = utils.to_hex_str(transaction_hash)

        self.transaction_hash = transaction_hash
        self.input = neo_express_data_file

        super().__init__('transaction', [transaction_hash])

    def _get_options(self) -> dict[str, str]:
        return super()._get_options()



import ast
from typing import Any

from boa3.internal.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.internal.model.expression import IExpression
from boa3.internal.model.type.itype import IType
from boa3.internal.model.type.primitive.primitivetype import PrimitiveType
from boa3.internal.model.variable import Variable


class ExceptionMethod(IBuiltinMethod):

    def __init__(self, argument_type: IType = None):
        from boa3.internal.model.type.type import Type
        if argument_type is None or not self.validate_parameters(argument_type):
            argument_type = Type.str

        identifier = '-Exception'
        args: dict[str, Variable] = {'message': Variable(argument_type)}
        default_message = "'{0}'".format(self.default_message) if argument_type is Type.str else "{0}"
        default = ast.parse(default_message.format(argument_type.default_value)
                            ).body[0].value
        super().__init__(identifier, args, [default], return_type=Type.exception)

    @property
    def _arg_message(self) -> Variable:
        return self.args['message']

    @property
    def default_message(self) -> str:
        return 'usererror'

    @property
    def identifier(self) -> str:
        return self._identifier

    def validate_parameters(self, *params: IExpression) -> bool:
        if len(params) > 1:
            return False
        if len(params) == 0:
            return True

        from boa3.internal.model.type.itype import IType
        if not isinstance(params[0], (IExpression, IType)):
            return False

        param_type: IType = params[0].type if isinstance(params[0], IExpression) else params[0]
        return isinstance(param_type, PrimitiveType)

    def generate_internal_opcodes(self, code_generator):
        pass

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> str | None:
        return

    def build(self, value: Any) -> IBuiltinMethod:
        if type(value) == type(self._arg_message.type):
            return self
        if isinstance(value, list):
            value = value[0] if len(value) > 0 else None
        if self.validate_parameters(value):
            return ExceptionMethod(value)
        return super().build(value)



from abc import ABC, abstractmethod
from typing import Self


class ISymbol(ABC):
    defined_by_entry: bool = False

    @property
    @abstractmethod
    def shadowing_name(self) -> str:
        """
        Gets the type of the evaluated expression

        :return: the resulting type when the expression is evaluated
        """
        pass

    @property
    @abstractmethod
    def is_deprecated(self) -> bool:
        """
        Whether the symbol is deprecated.
        """
        return False

    @abstractmethod
    def deprecate(self, new_location: str = None):
        """
        Deprecate this symbol
        """
        pass

    def clone(self) -> Self:
        obj = object.__new__(type(self))
        for key, value in self.__dict__.items():
            obj.__dict__[key] = value
        return obj



from boa3.internal.model.operation.binary.binaryoperation import BinaryOperation
from boa3.internal.model.operation.operator import Operator
from boa3.internal.model.type.collection.icollection import ICollectionType
from boa3.internal.model.type.collection.mapping.mappingtype import MappingType
from boa3.internal.model.type.primitive.primitivetype import PrimitiveType
from boa3.internal.model.type.type import IType, Type


class CollectionNotMembership(BinaryOperation):
    """
    A class used to represent a collection not membership operation

    :ivar operator: the operator of the operation. Inherited from :class:`IOperation`
    :ivar left: the left operand type. Inherited from :class:`BinaryOperation`
    :ivar right: the left operand type. Inherited from :class:`BinaryOperation`
    :ivar result: the result type of the operation.  Inherited from :class:`IOperation`
    """
    _valid_types: list[IType] = [Type.collection]

    def __init__(self, left: IType = Type.any, right: IType = None):
        self.operator: Operator = Operator.NotIn

        if not isinstance(right, ICollectionType):
            right = Type.collection

        if not right.item_type.is_type_of(left) or right.item_type is Type.any and left is not Type.any:
            right = Type.collection.build(Type.sequence.build(left))

        super().__init__(left, right)

    def validate_type(self, *types: IType) -> bool:
        if len(types) != self.number_of_operands:
            return False
        left: IType = types[0]
        right: IType = types[1]

        if not isinstance(right, ICollectionType):
            return False

        if isinstance(right, PrimitiveType) and (right.is_type_of(left) or left.is_type_of(right)):
            return True

        if isinstance(right, MappingType):
            reference_type = right.key_type
        else:
            reference_type = right.item_type

        return reference_type.is_type_of(left)

    def _get_result(self, left: IType, right: IType) -> IType:
        return Type.bool

    def get_valid_operand_for_validation(self, left_operand: IType,
                                         right_operand: IType = None) -> tuple[IType | None, IType | None]:
        if isinstance(right_operand, ICollectionType):
            left = right_operand.item_type if not isinstance(right_operand, MappingType) else right_operand.key_type
            return left, right_operand

        return left_operand, self.right_type

    def generate_internal_opcodes(self, code_generator):
        from boa3.internal.model.operation.binary.additional import CollectionMembership
        from boa3.internal.model.operation.unaryop import UnaryOp

        code_generator.convert_operation(CollectionMembership(self.left_type, self.right_type),
                                         is_internal=True)
        code_generator.convert_operation(UnaryOp.Not, is_internal=True)



from boa3.internal.model.builtin.interop.storage.put.istorageputmethod import IStoragePutMethod


class StoragePutUInt160Method(IStoragePutMethod):
    def __init__(self):
        from boa3.internal.model.type.collection.sequence.uint160type import UInt160Type

        identifier = 'put_uint160'
        value_type = UInt160Type.build()

        super().__init__(identifier, value_type=value_type)

    def generate_serialize_value_opcodes(self, code_generator):
        # it doesn't need to serialize UInt160 because it is also bytes value
        return



import ast
from typing import Any, Sized

from boa3.internal.model.builtin.method import IBuiltinMethod
from boa3.internal.model.builtin.native.stdlib.stringsplitmethod import \
    StringSplitWithoutRemoveEmptyEntriesMethod
from boa3.internal.model.type.type import Type
from boa3.internal.model.variable import Variable
from boa3.internal.neo.vm.opcode.Opcode import Opcode


class StrSplitMethod(IBuiltinMethod):
    def __init__(self, args: dict[str, Variable] = None, defaults: list[ast.AST] = None):
        identifier = 'split'

        if args is None:
            args: dict[str, Variable] = {
                'self': Variable(Type.str),
                'sep': Variable(Type.str),
                'maxsplit': Variable(Type.int)
            }

        if defaults is None:
            # whitespace is the default separator
            separator_default = ast.parse("' '").body[0].value
            # maxsplit the default value is -1
            maxsplit_default = ast.parse("-1").body[0].value.operand
            maxsplit_default.n = -1
            defaults = [separator_default, maxsplit_default]

        super().__init__(identifier, args, defaults,
                         return_type=Type.list.build_collection(Type.str))

    @property
    def identifier(self) -> str:
        if 1 <= len(self.args) <= 2:
            return '-{0}_2_args'.format(self._identifier)
        return self._identifier

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> str | None:
        return None

    @property
    def generation_order(self) -> list[int]:
        # the original string must be the top value in the stack
        indexes = list(range(len(self.args)))
        str_index = list(self.args).index('self')

        if indexes[-1] != str_index:
            # context must be the last generated argument
            indexes.remove(str_index)
            indexes.append(str_index)
        return indexes

    def generate_internal_opcodes(self, code_generator):
        from boa3.internal.model.builtin.builtin import Builtin
        from boa3.internal.model.operation.binaryop import BinaryOp

        code_generator.duplicate_stack_item(3)
        code_generator.swap_reverse_stack_items(2)

        code_generator.convert_builtin_method_call(StringSplitWithoutRemoveEmptyEntriesMethod())

        # if maxsplit > 0
        code_generator.duplicate_stack_item(2)
        code_generator.convert_literal(0)
        is_valid_split_count = code_generator.convert_begin_if()
        code_generator.change_jump(is_valid_split_count, Opcode.JMPLT)

        #   while len(array) < maxsplit + 1:
        while_start = code_generator.convert_begin_while()
        code_generator.duplicate_stack_top_item()

        #       concat values
        concat_operation = BinaryOp.Concat.build(Type.bytearray)  # to return buffer

        code_generator.duplicate_stack_item(4)
        code_generator.duplicate_stack_item(2)
        code_generator.insert_opcode(Opcode.POPITEM, pop_from_stack=True, add_to_stack=[Type.str])
        code_generator.convert_operation(concat_operation, is_internal=True)
        code_generator.duplicate_stack_item(2)
        code_generator.insert_opcode(Opcode.POPITEM, pop_from_stack=True, add_to_stack=[Type.str])
        code_generator.swap_reverse_stack_items(2)
        code_generator.convert_operation(concat_operation, is_internal=True)
        code_generator.convert_cast(Type.str, is_internal=True)
        code_generator.convert_builtin_method_call(Builtin.SequenceAppend, is_internal=True)

        while_condition = code_generator.bytecode_size
        code_generator.duplicate_stack_top_item()
        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)
        code_generator.duplicate_stack_item(3)
        code_generator.insert_opcode(Opcode.INC)
        code_generator.convert_operation(BinaryOp.Gt, is_internal=True)

        code_generator.convert_end_while(while_start, while_condition, is_internal=True)

        code_generator.convert_end_if(is_valid_split_count, is_internal=True)
        # clean stack

        code_generator.swap_reverse_stack_items(3)
        code_generator.remove_stack_top_item()
        code_generator.remove_stack_top_item()

    def build(self, value: Any) -> IBuiltinMethod:
        if isinstance(value, Sized) and 1 <= len(value) <= 2:
            return StrSplitWithoutMaxsplitMethod()

        return self


class StrSplitWithoutMaxsplitMethod(StrSplitMethod):
    def __init__(self):
        args: dict[str, Variable] = {
            'self': Variable(Type.str),
            'sep': Variable(Type.str),
        }
        # whitespace is the default separator
        separator_default = ast.parse("' '").body[0].value

        super().__init__(args, defaults=[separator_default])

    def generate_internal_opcodes(self, code_generator):
        code_generator.convert_builtin_method_call(StringSplitWithoutRemoveEmptyEntriesMethod())



def list_inner_packages(rootdir: str) -> list:
    import os.path

    inner_dirs = []

    if os.path.isdir(rootdir):
        rootdir = os.path.abspath(rootdir)
        for file in os.listdir(rootdir):
            dir_ = os.path.join(rootdir, file)

            if os.path.isdir(dir_):
                inner_dirs.extend(list_inner_packages(dir_))
                if os.path.isfile(os.path.join(dir_, '__init__.py')):
                    inner_dirs.append(dir_)

    return inner_dirs



from typing import Any

from boa3.internal.model.type.enum.intflagtype import IntFlagType
from boa3.internal.model.type.itype import IType
from boa3.internal.neo3.contracts.findoptions import FindOptions


class FindOptionsType(IntFlagType):
    """
    A class used to represent Neo interop FindOptions type
    """

    def __init__(self):
        super().__init__(FindOptions)
        self._identifier = 'FindOptions'

    @property
    def default_value(self) -> Any:
        return FindOptions.NONE

    @classmethod
    def build(cls, value: Any = None) -> IType:
        if cls._is_type_of(value) or value is None:
            from boa3.internal.model.builtin.interop.interop import Interop
            return Interop.FindOptionsType

    @classmethod
    def _is_type_of(cls, value: Any):
        return isinstance(value, (FindOptions, FindOptionsType))



from typing import Any

from boa3.internal.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.internal.model.expression import IExpression
from boa3.internal.model.type.collection.sequence.mutable.mutablesequencetype import MutableSequenceType
from boa3.internal.model.variable import Variable


class InsertMethod(IBuiltinMethod):
    def __init__(self, sequence_type: MutableSequenceType = None):
        if not isinstance(sequence_type, MutableSequenceType):
            from boa3.internal.model.type.type import Type
            sequence_type = Type.mutableSequence

        self_arg = Variable(sequence_type)
        index_arg = Variable(sequence_type.valid_key)
        item_arg = Variable(sequence_type.value_type)

        identifier = 'insert'
        args: dict[str, Variable] = {'self': self_arg,
                                     '__index': index_arg,
                                     '__object': item_arg}
        super().__init__(identifier, args)

    @property
    def _arg_self(self) -> Variable:
        return self.args['self']

    def validate_parameters(self, *params: IExpression) -> bool:
        if len(params) != 3:
            return False
        if not all(isinstance(param, IExpression) for param in params):
            return False

        from boa3.internal.model.type.itype import IType
        sequence_type: IType = params[0].type
        index_type: IType = params[1].type
        value_type: IType = params[2].type

        if not isinstance(sequence_type, MutableSequenceType):
            return False
        return sequence_type.key_type.is_type_of(index_type) and sequence_type.value_type.is_type_of(value_type)

    def validate_negative_arguments(self) -> list[int]:
        return [list(self.args).index('__index')]

    def generate_internal_opcodes(self, code_generator):
        from boa3.internal.model.builtin.builtin import Builtin
        from boa3.internal.model.operation.binaryop import BinaryOp
        from boa3.internal.neo.vm.opcode.Opcode import Opcode

        code_generator.duplicate_stack_item(3)
        code_generator.duplicate_stack_top_item()
        # array.append(None)
        code_generator.convert_literal(None)
        code_generator.convert_builtin_method_call(Builtin.SequenceAppend, is_internal=True)

        code_generator.duplicate_stack_top_item()
        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)
        code_generator.insert_opcode(Opcode.DEC)
        # x = len(array) - 1
        code_generator.duplicate_stack_top_item()
        code_generator.duplicate_stack_item(4)

        # if index > x:
        #   index = x
        if_index_gt_array = code_generator.convert_begin_if()
        code_generator.change_jump(if_index_gt_array, Opcode.JMPGT)
        code_generator.swap_reverse_stack_items(3)
        code_generator.remove_stack_top_item()
        code_generator.duplicate_stack_item(2)
        else_index_le_array = code_generator.convert_begin_else(if_index_gt_array, is_internal=True)

        code_generator.convert_literal(0)
        code_generator.duplicate_stack_item(4)
        # elif index < 0:
        #  index = 0
        elif_index_lt_zero = code_generator.convert_begin_if()
        code_generator.change_jump(elif_index_lt_zero, Opcode.JMPLE)
        code_generator.swap_reverse_stack_items(3)
        code_generator.remove_stack_top_item()
        code_generator.convert_literal(0)
        code_generator.swap_reverse_stack_items(3)
        code_generator.convert_end_if(elif_index_lt_zero, is_internal=True)
        code_generator.convert_end_if(else_index_le_array, is_internal=True)

        # while x > index:
        while_begin = code_generator.convert_begin_while()

        #   array[x] = array[x - 1]
        code_generator.duplicate_stack_top_item()
        code_generator.insert_opcode(Opcode.DEC)

        code_generator.duplicate_stack_item(3)
        code_generator.duplicate_stack_top_item()
        code_generator.swap_reverse_stack_items(3)
        code_generator.convert_get_item(index_inserted_internally=True)

        code_generator.duplicate_stack_item(3)
        value_address = code_generator.bytecode_size
        code_generator.swap_reverse_stack_items(2)
        code_generator.convert_set_item(value_address, index_inserted_internally=True)

        #   x =- 1
        code_generator.insert_opcode(Opcode.DEC)

        while_condition = code_generator.bytecode_size
        code_generator.duplicate_stack_top_item()
        code_generator.duplicate_stack_item(4)
        code_generator.convert_operation(BinaryOp.Gt)

        code_generator.convert_end_while(while_begin, while_condition, is_internal=True)

        # clean stack from loop variables
        code_generator.remove_stack_top_item()
        code_generator.remove_stack_top_item()

        # array[index] = object
        value_address = code_generator.bytecode_size
        code_generator.swap_reverse_stack_items(2)
        code_generator.convert_set_item(value_address, index_inserted_internally=True)

    def push_self_first(self) -> bool:
        return self.has_self_argument

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> str | None:
        return None

    def build(self, value: Any) -> IBuiltinMethod:
        if type(value) == type(self.args['self'].type):
            return self
        if isinstance(value, MutableSequenceType):
            return InsertMethod(value)
        return super().build(value)
